"""
Module to provide processing for the fenced leaf blocks.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.container_blocks.container_grab_bag import ContainerGrabBag
from pymarkdown.container_blocks.container_helper import ContainerHelper
from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.inline.inline_backslash_helper import InlineBackslashHelper
from pymarkdown.leaf_blocks.leaf_block_helper import LeafBlockHelper
from pymarkdown.tokens.fenced_code_block_markdown_token import (
    FencedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.tokens.stack_token import (
    FencedCodeBlockStackToken,
    ListStackToken,
    ParagraphStackToken,
    StackToken,
)
from pymarkdown.tokens.text_markdown_token import TextMarkdownToken

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-many-lines


class FencedLeafBlockProcessor:
    """
    Class to provide processing for the fenced leaf blocks.
    """

    __fenced_start_tilde = "~"
    __fenced_start_backtick = "`"
    __fenced_code_block_start_characters = (
        f"{__fenced_start_tilde}{__fenced_start_backtick}"
    )

    # pylint: disable=too-many-arguments
    @staticmethod
    def is_fenced_code_block(
        parser_state: ParserState,
        line_to_parse: str,
        start_index: int,
        extracted_whitespace: str,
        original_line: str,
        index_indent: int,
        skip_whitespace_check: bool = False,
    ) -> Tuple[bool, Optional[int], Optional[int], Optional[int], Optional[int]]:
        """
        Determine if we have the start of a fenced code block.
        """

        after_fence_index: Optional[int] = None
        position_marker = PositionMarker(0, 0, "", index_indent)
        extracted_whitespace = LeafBlockHelper.realize_leading_whitespace(
            parser_state, position_marker, extracted_whitespace, original_line
        )

        if (
            skip_whitespace_check
            or TabHelper.is_length_less_than_or_equal_to(extracted_whitespace, 3)
        ) and ParserHelper.is_character_at_index_one_of(
            line_to_parse,
            start_index,
            FencedLeafBlockProcessor.__fenced_code_block_start_characters,
        ):
            POGGER.debug("ifcb:collected_count>:$:<$<<", line_to_parse, start_index)
            collected_count, new_index = ParserHelper.collect_while_character_verified(
                line_to_parse, start_index, line_to_parse[start_index]
            )
            POGGER.debug(
                "ifcb:collected_count:$, new_index:$", collected_count, new_index
            )
            after_fence_index = new_index
            (
                non_whitespace_index,
                _,
            ) = ParserHelper.extract_ascii_whitespace(line_to_parse, new_index)

            if collected_count >= 3:
                POGGER.debug("ifcb:True")
                POGGER.debug("ifcb:collected_count=$", collected_count)
                POGGER.debug("ifcb:non_whitespace_index=$", non_whitespace_index)
                POGGER.debug("ifcb:after_fence_index=$", after_fence_index)
                return (
                    True,
                    non_whitespace_index,
                    after_fence_index,
                    collected_count,
                    new_index,
                )
        return False, None, None, None, None

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def parse_fenced_code_block(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        original_line: str,
        detabified_original_start_index: int,
        block_quote_data: BlockQuoteData,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[List[MarkdownToken], str]:
        """
        Handle the parsing of a fenced code block
        """

        POGGER.debug(
            "line>>$>>index>>$>>",
            position_marker.text_to_parse,
            position_marker.index_number,
        )
        POGGER.debug(
            "parse_fenced_code_block:extracted_whitespace-->$<--", extracted_whitespace
        )
        new_tokens: List[MarkdownToken] = []
        (
            is_fence_start,
            non_whitespace_index,
            after_fence_index,
            collected_count,
            new_index,
        ) = FencedLeafBlockProcessor.is_fenced_code_block(
            parser_state,
            position_marker.text_to_parse,
            position_marker.index_number,
            extracted_whitespace,
            original_line,
            position_marker.index_indent,
        )
        if is_fence_start and not parser_state.token_stack[-1].is_html_block:
            POGGER.debug("parse_fenced_code_block:fenced")
            assert (
                non_whitespace_index is not None
                and after_fence_index is not None
                and collected_count is not None
                and new_index is not None
            ), "If is_fence_start, all these must be defined."
            if parser_state.token_stack[-1].is_fenced_code_block:
                POGGER.debug("parse_fenced_code_block:check fence end")
                FencedLeafBlockProcessor.__check_for_fenced_end(
                    parser_state,
                    position_marker,
                    collected_count,
                    extracted_whitespace,
                    new_tokens,
                    after_fence_index,
                    original_line,
                    detabified_original_start_index,
                )
            else:
                POGGER.debug("parse_fenced_code_block:check fence start")
                new_tokens = FencedLeafBlockProcessor.__process_fenced_start(
                    parser_state,
                    position_marker,
                    non_whitespace_index,
                    collected_count,
                    extracted_whitespace,
                    original_line,
                    new_index,
                    block_quote_data,
                    after_fence_index,
                    grab_bag,
                )
        elif parser_state.token_stack[-1].is_fenced_code_block:
            POGGER.debug("parse_fenced_code_block:already in")
            extracted_whitespace = (
                FencedLeafBlockProcessor.__parse_fenced_code_block_already_in(
                    parser_state,
                    extracted_whitespace,
                    original_line,
                    position_marker.text_to_parse,
                    grab_bag.indent_used_by_list,
                )
            )
            POGGER.debug("parse_fenced_code_block:already in-->$", extracted_whitespace)
        return new_tokens, extracted_whitespace

    # pylint: enable=too-many-arguments

    @staticmethod
    def __check_for_fenced_end_with_split_tab(
        parser_state: ParserState, position_marker: PositionMarker, original_line: str
    ) -> Tuple[bool, str]:
        reconstructed_line = position_marker.text_to_parse
        was_indented = not parser_state.token_stack[-2].is_document
        if is_bq_start := original_line.startswith(">"):
            indent_prefix: Optional[str] = None
        else:
            indent_prefix = " " * position_marker.index_indent
        (
            _,
            adj_original_index,
            split_tab,
        ) = TabHelper.find_tabified_string(
            original_line,
            reconstructed_line,
            abc=is_bq_start,
            was_indented=was_indented,
            reconstruct_prefix=indent_prefix,
        )
        split_tab_whitespace = original_line[:adj_original_index]

        return split_tab, split_tab_whitespace

    # pylint: disable=too-many-arguments
    @staticmethod
    def __check_for_fenced_end(
        parser_state: ParserState,
        position_marker: PositionMarker,
        collected_count: int,
        extracted_whitespace: str,
        new_tokens: List[MarkdownToken],
        after_fence_index: int,
        original_line: str,
        detabified_original_start_index: int,
    ) -> None:
        # POGGER.debug("pfcb->end")
        # POGGER.debug("extracted_whitespace:$:", extracted_whitespace)
        # POGGER.debug("position_marker.text_to_parse:$:", position_marker.text_to_parse)
        # POGGER.debug(
        #     "len(position_marker.text_to_parse):$:", len(position_marker.text_to_parse)
        # )
        # POGGER.debug("after_fence_index:$:", after_fence_index)
        # POGGER.debug(
        #     "detabified_original_start_index:$:", detabified_original_start_index
        # )

        # POGGER.debug("original_line:$:", original_line)
        only_spaces_after_fence = True
        split_tab = False
        split_tab_whitespace: Optional[str] = None
        if ParserHelper.tab_character in original_line:
            (
                after_fence_index,
                only_spaces_after_fence,
                extracted_whitespace,
                split_tab,
            ) = FencedLeafBlockProcessor.__check_for_fenced_end_with_tab(
                parser_state,
                position_marker,
                original_line,
                detabified_original_start_index,
                collected_count,
                extracted_whitespace,
            )
            if split_tab:
                (
                    split_tab,
                    split_tab_whitespace,
                ) = FencedLeafBlockProcessor.__check_for_fenced_end_with_split_tab(
                    parser_state, position_marker, original_line
                )

        after_fence_and_spaces_index, extracted_spaces = (
            ParserHelper.extract_spaces_verified(
                position_marker.text_to_parse, after_fence_index
            )
        )
        # POGGER.debug("after_fence_and_spaces_index:$:", after_fence_and_spaces_index)

        fenced_token = cast(FencedCodeBlockStackToken, parser_state.token_stack[-1])
        # POGGER.debug("fenced_token.code_fence_character:$:", fenced_token.code_fence_character)
        # POGGER.debug("position_marker.text_to_parse[position_marker.index_number]:$:", position_marker.text_to_parse[position_marker.index_number])
        # POGGER.debug("collected_count:$:", collected_count)
        # POGGER.debug("fenced_token.fence_character_count:$:", fenced_token.fence_character_count)
        # POGGER.debug("after_fence_and_spaces_index:$:", after_fence_and_spaces_index)
        # POGGER.debug("len(position_marker.text_to_parse):$:", len(position_marker.text_to_parse))
        # POGGER.debug("only_spaces_after_fence:$:", only_spaces_after_fence)

        if (
            fenced_token.code_fence_character
            == position_marker.text_to_parse[position_marker.index_number]
            and collected_count >= fenced_token.fence_character_count
            and after_fence_and_spaces_index >= len(position_marker.text_to_parse)
            and only_spaces_after_fence
        ):
            if split_tab:
                TabHelper.adjust_block_quote_indent_for_tab(
                    parser_state, split_tab_whitespace
                )

            new_end_token = parser_state.token_stack[
                -1
            ].generate_close_markdown_token_from_stack_token(
                extracted_whitespace,
                extra_end_data=f"{extracted_spaces}:{collected_count}",
            )

            new_tokens.append(new_end_token)
            del parser_state.token_stack[-1]

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __check_for_fenced_end_with_tab(
        parser_state: ParserState,
        position_marker: PositionMarker,
        original_line: str,
        detabified_original_start_index: int,
        collected_count: int,
        extracted_whitespace: str,
    ) -> Tuple[int, bool, str, bool]:
        (
            after_fence_index,
            adj_end,
            fence_string,
        ) = FencedLeafBlockProcessor.__calculate_fenced_vars(
            collected_count, original_line, detabified_original_start_index
        )

        original_fence_string_index = original_line.find(fence_string)
        assert (
            original_fence_string_index != -1
        ), "fence_string must be in original line"
        after_fence_in_original = original_line[
            original_fence_string_index + collected_count :
        ]

        _, after_space_in_original_index = ParserHelper.collect_while_character(
            after_fence_in_original, 0, " "
        )
        only_spaces_after_fence = after_space_in_original_index == len(
            after_fence_in_original
        )

        was_indented = not parser_state.token_stack[-2].is_document
        indent_prefix: Optional[str] = (
            None
            if original_line.startswith(">")
            else " " * position_marker.index_indent
        )

        _, adj_original_index, split_tab = TabHelper.find_tabified_string(
            original_line,
            extracted_whitespace + adj_end,
            was_indented=was_indented,
            reconstruct_prefix=indent_prefix,
        )

        _, new_extracted_whitespace = ParserHelper.extract_spaces_verified(
            original_line, adj_original_index
        )

        return (
            after_fence_index,
            only_spaces_after_fence,
            new_extracted_whitespace,
            split_tab,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __calculate_fenced_vars(
        collected_count: int, original_line: str, detabified_original_start_index: int
    ) -> Tuple[int, str, str]:
        detabified_original_line = TabHelper.detabify_string(original_line)
        adj_original_line = detabified_original_line[detabified_original_start_index:]

        after_whitespace_index, _ = ParserHelper.extract_spaces_verified(
            adj_original_line, 0
        )
        after_fence_index = after_whitespace_index + collected_count
        adj_end = adj_original_line[after_whitespace_index:]
        fence_string = adj_original_line[after_whitespace_index:after_fence_index]
        assert fence_string in original_line, "fence_string must be in original line"
        return after_fence_index, adj_end, fence_string

    @staticmethod
    def __process_fenced_start_adjust(
        parser_state: ParserState,
        position_marker: PositionMarker,
        new_tokens: List[MarkdownToken],
    ) -> None:
        pre_fenced_block_index = len(parser_state.token_stack) - 2
        if parser_state.token_stack[pre_fenced_block_index].is_list:
            additional_column_delta = (
                cast(
                    ListStackToken, parser_state.token_stack[pre_fenced_block_index]
                ).indent_level
                - position_marker.index_indent
            )
        else:
            additional_column_delta = 0
        fenced_token = cast(FencedCodeBlockMarkdownToken, new_tokens[-1])
        assert fenced_token.is_fenced_code_block
        new_position_marker = PositionMarker(
            line_number=fenced_token.line_number,
            index_number=fenced_token.column_number - 1 + additional_column_delta,
            text_to_parse="",
            index_indent=0,
        )
        new_fenced_token = FencedCodeBlockMarkdownToken(
            fence_character=fenced_token.fence_character,
            fence_count=fenced_token.fence_count,
            extracted_text=fenced_token.extracted_text,
            pre_extracted_text=fenced_token.pre_extracted_text,
            text_after_extracted_text=fenced_token.text_after_extracted_text,
            pre_text_after_extracted_text=fenced_token.pre_text_after_extracted_text,
            extracted_whitespace=fenced_token.extracted_whitespace,
            extracted_whitespace_before_info_string=fenced_token.extracted_whitespace_before_info_string,
            position_marker=new_position_marker,
        )
        fenced_stack_token = parser_state.token_stack[-1]
        fenced_stack_token.reset_matching_markdown_token(new_fenced_token)
        adjusted_token_list = new_tokens[:-1]
        adjusted_token_list.append(new_fenced_token)
        new_tokens.clear()
        new_tokens.extend(adjusted_token_list)

    # pylint: disable=too-many-boolean-expressions
    @staticmethod
    def __process_fenced_start_kludge(
        parser_state: ParserState,
        position_marker: PositionMarker,
        new_tokens: List[MarkdownToken],
    ) -> None:
        if (
            len(parser_state.token_stack) == 3
            and parser_state.token_stack[-2].is_list
            and len(new_tokens) == 4
            and new_tokens[-2].is_block_quote_end
            and new_tokens[-3].is_list_end
            and new_tokens[-4].is_block_quote_end
        ):
            list_end_token = cast(EndMarkdownToken, new_tokens[-3])
            list_start_token = cast(
                ListStartMarkdownToken, list_end_token.start_markdown_token
            )
            assert list_start_token.leading_spaces is not None
            leading_space_split = list_start_token.leading_spaces.split("\n")
            assert len(leading_space_split) >= 2

            # Looks a bit weird, but we need to adjust the second last leading space.
            POGGER.debug("__process_fenced_start>>list_token>>$", list_start_token)
            POGGER.debug(
                "__process_fenced_start_kludge>>list_start_token>>$",
                list_start_token,
            )
            last_leading_space = list_start_token.remove_last_leading_space()
            POGGER.debug(
                "__process_fenced_start_kludge>>list_start_token>>$",
                list_start_token,
            )
            assert last_leading_space is not None
            POGGER.debug(
                "__process_fenced_start_kludge>>list_start_token>>$",
                list_start_token,
            )
            list_start_token.remove_last_leading_space()
            POGGER.debug(
                "__process_fenced_start_kludge>>list_start_token>>$",
                list_start_token,
            )
            POGGER.debug("__process_fenced_start>>list_token>>$", list_start_token)
            list_start_token.add_leading_spaces(" " * position_marker.index_indent)
            list_start_token.add_leading_spaces(last_leading_space)
            POGGER.debug("__process_fenced_start>>list_token>>$", list_start_token)

    # pylint: enable=too-many-boolean-expressions

    # pylint: disable=too-many-arguments
    @staticmethod
    def __process_fenced_start(
        parser_state: ParserState,
        position_marker: PositionMarker,
        non_whitespace_index: int,
        collected_count: int,
        extracted_whitespace: str,
        original_line: str,
        new_index: int,
        block_quote_data: BlockQuoteData,
        after_fence_index: int,
        grab_bag: ContainerGrabBag,
    ) -> List[MarkdownToken]:
        POGGER.debug("pfcb->check")
        new_tokens: List[MarkdownToken] = []
        if (
            position_marker.text_to_parse[position_marker.index_number]
            == FencedLeafBlockProcessor.__fenced_start_tilde
            or FencedLeafBlockProcessor.__fenced_start_backtick
            not in position_marker.text_to_parse[non_whitespace_index:]
        ):
            POGGER.debug("pfcb->start")

            (
                old_top_of_stack,
                new_tokens,
                adjusted_corrected_prefix,
            ) = FencedLeafBlockProcessor.__add_fenced_tokens(
                parser_state,
                position_marker,
                non_whitespace_index,
                collected_count,
                extracted_whitespace,
                original_line,
                new_index,
                block_quote_data,
                after_fence_index,
                grab_bag,
            )

            POGGER.debug("new_tokens-->$<<", new_tokens)
            POGGER.debug("StackToken-->$<<", parser_state.token_stack[-1])
            POGGER.debug(
                "StackToken>start_markdown_token-->$<<",
                parser_state.token_stack[-1].matching_markdown_token,
            )
            FencedLeafBlockProcessor.__process_fenced_start_kludge(
                parser_state, position_marker, new_tokens
            )

            removed_char_length = (
                len(adjusted_corrected_prefix)
                if adjusted_corrected_prefix is not None
                else None
            )
            before_correct_length = len(new_tokens)
            LeafBlockHelper.correct_for_leaf_block_start_in_list(
                parser_state,
                position_marker.index_indent,
                old_top_of_stack,
                new_tokens,
                block_quote_data,
                alt_removed_chars_at_start=removed_char_length,
                original_line=original_line,
            )

            # TODO restructure this so we do not have to do this kludge
            # need to place the correct_for_leaf_block_start_in_list call before
            # we create the fenced token
            POGGER.debug("new_tokens-->$<<", new_tokens)
            if before_correct_length != len(new_tokens):
                FencedLeafBlockProcessor.__process_fenced_start_adjust(
                    parser_state, position_marker, new_tokens
                )
        return new_tokens

    # pylint: enable=too-many-arguments

    @staticmethod
    def __parse_fenced_code_block_already_in(
        parser_state: ParserState,
        extracted_whitespace: str,
        original_line: str,
        line_to_parse: str,
        indent_used_by_list: Optional[str],
    ) -> str:
        fenced_token = cast(FencedCodeBlockStackToken, parser_state.token_stack[-1])
        if fenced_token.whitespace_start_count and extracted_whitespace:
            POGGER.debug("original_line>:$:<", original_line)
            POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
            current_whitespace_length = TabHelper.calculate_length(extracted_whitespace)
            POGGER.debug("current_whitespace_length>>$", current_whitespace_length)
            whitespace_left_count = max(
                0,
                current_whitespace_length - fenced_token.whitespace_start_count,
            )
            whitespace_used_count = current_whitespace_length - whitespace_left_count
            POGGER.debug("previous_ws>>$", current_whitespace_length)
            POGGER.debug("whitespace_left>:$:<", whitespace_left_count)
            (
                do_normal_processing,
                removed_whitespace,
                whitespace_padding,
            ) = FencedLeafBlockProcessor.__parse_fenced_code_block_already_in_with_tab(
                parser_state,
                whitespace_used_count,
                whitespace_left_count,
                original_line,
                line_to_parse,
                current_whitespace_length,
                indent_used_by_list,
            )
            if do_normal_processing:
                removed_whitespace = ParserHelper.create_replace_with_nothing_marker(
                    ParserHelper.repeat_string(
                        ParserHelper.space_character,
                        current_whitespace_length - whitespace_left_count,
                    )
                )
                whitespace_padding = ParserHelper.repeat_string(
                    ParserHelper.space_character, whitespace_left_count
                )
            extracted_whitespace = f"{removed_whitespace}{whitespace_padding}"
        POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
        return extracted_whitespace

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __parse_fenced_code_block_already_in_with_tab(
        parser_state: ParserState,
        whitespace_used_count: int,
        whitespace_left_count: int,
        original_line: str,
        line_to_parse: str,
        current_whitespace_length: int,
        indent_used_by_list: Optional[str],
    ) -> Tuple[bool, str, str]:
        do_normal_processing = True
        removed_whitespace = ""
        whitespace_padding = ""
        if whitespace_left_count and "\t" in original_line:
            if original_line.startswith("\t\t"):
                whitespace_padding = "\t"
                removed_whitespace = ParserHelper.create_replace_with_nothing_marker(
                    "\t"
                )
                do_normal_processing = False

                last_container_token = parser_state.token_stack[-2]
                assert last_container_token.is_list
                list_stack_token = cast(ListStackToken, last_container_token)
                list_markdown_token = cast(
                    ListStartMarkdownToken, list_stack_token.matching_markdown_token
                )
                POGGER.debug(
                    "__parse_fenced_code_block_already_in_with_tab>>list_markdown_token>>$",
                    list_markdown_token,
                )
                list_markdown_token.remove_last_leading_space()
                POGGER.debug(
                    "__parse_fenced_code_block_already_in_with_tab>>list_markdown_token>>$",
                    list_markdown_token,
                )
            else:
                ex_space, ex_space_index, split_tab = TabHelper.find_tabified_string(
                    original_line,
                    line_to_parse,
                    use_proper_traverse=True,
                    reconstruct_prefix=indent_used_by_list,
                )
                modified_ex_space_index = (
                    ex_space_index + 1 if split_tab else ex_space_index
                )

                (last_good_space_index, _, detabified_ex_space) = (
                    TabHelper.search_for_tabbed_prefix(
                        ex_space, modified_ex_space_index, whitespace_used_count
                    )
                )

                do_normal_processing = len(detabified_ex_space) == whitespace_used_count
                if not do_normal_processing:
                    (
                        removed_whitespace,
                        whitespace_padding,
                    ) = FencedLeafBlockProcessor.__parse_fenced_code_block_already_in_with_tab_whitespace(
                        ex_space,
                        last_good_space_index,
                        detabified_ex_space,
                        whitespace_used_count,
                        current_whitespace_length,
                    )
        return do_normal_processing, removed_whitespace, whitespace_padding

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __parse_fenced_code_block_already_in_with_tab_whitespace(
        ex_space: str,
        last_good_space_index: int,
        detabified_ex_space: str,
        whitespace_used_count: int,
        current_whitespace_length: int,
    ) -> Tuple[str, str]:
        tab_prefix = ex_space[: last_good_space_index - 1]
        split_tab_character = ex_space[last_good_space_index - 1]
        whitespace_left = detabified_ex_space[whitespace_used_count:]

        removed_whitespace = (
            ParserHelper.create_replace_with_nothing_marker(tab_prefix)
            if tab_prefix
            else ""
        )
        removed_whitespace += ParserHelper.create_replacement_markers(
            split_tab_character, whitespace_left
        )
        whitespace_padding = ParserHelper.repeat_string(
            ParserHelper.space_character,
            current_whitespace_length - len(detabified_ex_space),
        )
        return removed_whitespace, whitespace_padding

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __add_fenced_tokens(
        parser_state: ParserState,
        position_marker: PositionMarker,
        non_whitespace_index: int,
        collected_count: int,
        extracted_whitespace: str,
        original_line: str,
        new_index: int,
        block_quote_data: BlockQuoteData,
        after_fence_index: int,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[StackToken, List[MarkdownToken], Optional[str]]:
        (
            line_to_parse,
            non_whitespace_index,
            extracted_whitespace_before_info_string,
            extracted_whitespace,
            split_tab,
            corrected_prefix_length,
            extracted_text,
            text_after_extracted_text,
            split_tab_whitespace,
            adjusted_corrected_prefix,
        ) = FencedLeafBlockProcessor.__add_fenced_tokens_prepare(
            position_marker,
            original_line,
            new_index,
            non_whitespace_index,
            collected_count,
            extracted_whitespace,
            after_fence_index,
        )

        return FencedLeafBlockProcessor.__add_fenced_tokens_create(
            parser_state,
            position_marker,
            extracted_whitespace,
            extracted_whitespace_before_info_string,
            corrected_prefix_length,
            collected_count,
            extracted_text,
            text_after_extracted_text,
            line_to_parse,
            split_tab,
            block_quote_data,
            split_tab_whitespace,
            adjusted_corrected_prefix,
            grab_bag,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __add_fenced_tokens_calc(
        parser_state: ParserState,
        position_marker: PositionMarker,
        split_tab: bool,
        block_quote_data: BlockQuoteData,
        split_tab_whitespace: Optional[str],
        extracted_whitespace: str,
        grab_bag: ContainerGrabBag,
    ) -> Tuple[StackToken, List[MarkdownToken], int, str, Optional[str]]:
        old_top_of_stack = parser_state.token_stack[-1]
        new_tokens, _ = parser_state.close_open_blocks_fn(
            parser_state,
            only_these_blocks=[ParagraphStackToken],
        )

        split_tab, extracted_whitespace, whitespace_prefix = (
            ContainerHelper.reduce_containers_if_required(
                parser_state,
                position_marker,
                block_quote_data,
                new_tokens,
                split_tab,
                extracted_whitespace,
                grab_bag,
            )
        )
        if split_tab:
            TabHelper.adjust_block_quote_indent_for_tab(
                parser_state,
                extracted_whitespace=split_tab_whitespace,
                fenced_switch_enabled=True,
            )
            whitespace_count_delta = -1
        else:
            whitespace_count_delta = 0

        return (
            old_top_of_stack,
            new_tokens,
            whitespace_count_delta,
            extracted_whitespace,
            whitespace_prefix,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-locals, too-many-arguments
    @staticmethod
    def __add_fenced_tokens_prepare(
        position_marker: PositionMarker,
        original_line: str,
        new_index: int,
        non_whitespace_index: int,
        collected_count: int,
        extracted_whitespace: str,
        after_fence_index: int,
    ) -> Tuple[
        str,
        int,
        str,
        str,
        bool,
        int,
        str,
        str,
        Optional[str],
        Optional[str],
    ]:
        (
            line_to_parse,
            non_whitespace_index,
            extracted_whitespace_before_info_string,
            extracted_whitespace,
            split_tab,
            corrected_prefix_length,
            split_tab_whitespace,
            adjusted_corrected_prefix,
        ) = FencedLeafBlockProcessor.__add_fenced_tokens_with_tab(
            position_marker,
            original_line,
            new_index,
            non_whitespace_index,
            collected_count,
            extracted_whitespace,
            after_fence_index,
        )

        (
            _,
            proper_end_index,
        ) = ParserHelper.collect_backwards_while_one_of_characters(
            line_to_parse, -1, Constants.ascii_whitespace
        )
        adjusted_string = line_to_parse[:proper_end_index]
        non_whitespace_index = min(non_whitespace_index, len(adjusted_string))

        (
            after_extracted_text_index,
            extracted_text,
        ) = ParserHelper.extract_until_spaces_verified(
            adjusted_string, non_whitespace_index
        )
        return (
            line_to_parse,
            non_whitespace_index,
            extracted_whitespace_before_info_string,
            extracted_whitespace,
            split_tab,
            corrected_prefix_length,
            extracted_text,
            line_to_parse[after_extracted_text_index:],
            split_tab_whitespace,
            adjusted_corrected_prefix,
        )

    # pylint: enable=too-many-locals, too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __add_fenced_tokens_create(
        parser_state: ParserState,
        position_marker: PositionMarker,
        extracted_whitespace: str,
        extracted_whitespace_before_info_string: str,
        corrected_prefix_length: int,
        collected_count: int,
        extracted_text: str,
        text_after_extracted_text: str,
        line_to_parse: str,
        split_tab: bool,
        block_quote_data: BlockQuoteData,
        split_tab_whitespace: Optional[str],
        adjusted_corrected_prefix: Optional[str],
        grab_bag: ContainerGrabBag,
    ) -> Tuple[StackToken, List[MarkdownToken], Optional[str]]:
        (
            old_top_of_stack,
            new_tokens,
            whitespace_start_count,
            extracted_whitespace,
            whitespace_prefix,
        ) = FencedLeafBlockProcessor.__add_fenced_tokens_calc(
            parser_state,
            position_marker,
            split_tab,
            block_quote_data,
            split_tab_whitespace,
            extracted_whitespace,
            grab_bag,
        )

        pre_extracted_text, pre_text_after_extracted_text = (
            extracted_text,
            text_after_extracted_text,
        )

        extracted_text = InlineBackslashHelper.handle_backslashes(
            parser_state.parse_properties, extracted_text
        )
        text_after_extracted_text = InlineBackslashHelper.handle_backslashes(
            parser_state.parse_properties, text_after_extracted_text
        )

        if pre_extracted_text == extracted_text:
            pre_extracted_text = ""
        if pre_text_after_extracted_text == text_after_extracted_text:
            pre_text_after_extracted_text = ""
        if whitespace_prefix:
            extracted_whitespace = whitespace_prefix + extracted_whitespace

        new_token = FencedCodeBlockMarkdownToken(
            position_marker.text_to_parse[position_marker.index_number],
            collected_count,
            extracted_text,
            pre_extracted_text,
            text_after_extracted_text,
            pre_text_after_extracted_text,
            extracted_whitespace,
            extracted_whitespace_before_info_string,
            position_marker,
        )
        new_tokens.append(new_token)
        whitespace_start_count += (
            TabHelper.calculate_length(split_tab_whitespace, 0)
            if split_tab_whitespace is not None
            else TabHelper.calculate_length(
                extracted_whitespace, corrected_prefix_length
            )
        )
        parser_state.token_stack.append(
            FencedCodeBlockStackToken(
                code_fence_character=line_to_parse[position_marker.index_number],
                fence_character_count=collected_count,
                whitespace_start_count=whitespace_start_count,
                matching_markdown_token=new_token,
            )
        )
        return old_top_of_stack, new_tokens, adjusted_corrected_prefix

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __add_fenced_tokens_with_tab(
        position_marker: PositionMarker,
        original_line: str,
        new_index: int,
        non_whitespace_index: int,
        collected_count: int,
        extracted_whitespace: str,
        after_fence_index: int,
    ) -> Tuple[str, int, str, str, bool, int, Optional[str], Optional[str]]:
        split_tab = False
        corrected_prefix_length = 0
        line_to_parse = position_marker.text_to_parse
        split_tab_whitespace: Optional[str] = None
        adjusted_corrected_prefix: Optional[str] = None
        if ParserHelper.tab_character in original_line:
            (
                fence_string,
                adj_original_line,
                split_tab,
                extracted_whitespace,
                corrected_prefix_length,
                split_tab_whitespace,
                adjusted_corrected_prefix,
            ) = FencedLeafBlockProcessor.__add_fenced_tokens_with_tab_calc(
                original_line,
                line_to_parse,
                new_index,
                collected_count,
                after_fence_index,
                extracted_whitespace,
            )

            line_to_parse = adj_original_line
            new_index = line_to_parse.find(fence_string)
            new_index += len(fence_string)

            (
                non_whitespace_index,
                extracted_whitespace_before_info_string,
            ) = ParserHelper.extract_ascii_whitespace_verified(line_to_parse, new_index)
        else:
            (
                _,
                extracted_whitespace_before_info_string,
            ) = ParserHelper.extract_ascii_whitespace_verified(line_to_parse, new_index)
        return (
            line_to_parse,
            non_whitespace_index,
            extracted_whitespace_before_info_string,
            extracted_whitespace,
            split_tab,
            corrected_prefix_length,
            split_tab_whitespace,
            adjusted_corrected_prefix,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __add_fenced_tokens_with_tab_calc(
        original_line: str,
        line_to_parse: str,
        new_index: int,
        collected_count: int,
        after_fence_index: int,
        extracted_whitespace: str,
    ) -> Tuple[str, str, bool, str, int, Optional[str], Optional[str]]:
        fence_string = line_to_parse[new_index - collected_count : new_index]
        split_tab = False
        corrected_prefix_length = 0
        split_tab_whitespace: Optional[str] = None
        adjusted_corrected_prefix: Optional[str] = None

        adj_original_line, _, _ = TabHelper.find_detabify_string(
            original_line, line_to_parse, use_proper_traverse=True
        )

        if adj_original_line is None:
            adj_original_line = (
                FencedLeafBlockProcessor.__add_fenced_tokens_with_tab_calc_prefix(
                    line_to_parse,
                    original_line,
                    new_index,
                    collected_count,
                    after_fence_index,
                )
            )

        fence_string_index = original_line.find(fence_string)
        assert fence_string_index != -1, "fence string must be in original line"
        if prefix := original_line[:fence_string_index]:
            (
                corrected_prefix,
                corrected_suffix,
                split_tab,
                _,
            ) = TabHelper.match_tabbed_whitespace(extracted_whitespace, prefix)
            extracted_whitespace = corrected_suffix
            corrected_prefix_length = len(corrected_prefix)
            if split_tab:
                split_tab_whitespace = corrected_prefix + corrected_suffix
                adjusted_corrected_prefix = corrected_prefix
        return (
            fence_string,
            adj_original_line,
            split_tab,
            extracted_whitespace,
            corrected_prefix_length,
            split_tab_whitespace,
            adjusted_corrected_prefix,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __add_fenced_tokens_with_tab_calc_prefix(
        line_to_parse: str,
        original_line: str,
        new_index: int,
        collected_count: int,
        after_fence_index: int,
    ) -> str:
        line_suffix = line_to_parse[new_index - collected_count : after_fence_index]
        line_suffix_index = original_line.find(line_suffix)
        assert line_suffix_index != -1, "line_suffix must be in original_line"
        return original_line[line_suffix_index:]

    # pylint: disable=too-many-arguments
    @staticmethod
    def handle_fenced_code_block(
        parser_state: ParserState,
        position_marker: PositionMarker,
        leaf_token_whitespace: str,
        new_tokens: List[MarkdownToken],
        original_line: str,
        detabified_original_start_index: int,
        block_quote_data: BlockQuoteData,
        grab_bag: ContainerGrabBag,
    ) -> bool:
        """
        Take care of the processing for fenced code blocks.
        """
        if (
            parser_state.token_stack[-1].was_link_definition_started
            or parser_state.token_stack[-1].was_table_block_started
        ):
            return False

        POGGER.debug(">>__handle_fenced_code_block>>start")
        POGGER.debug(">>leaf_token_whitespace>:$:<", leaf_token_whitespace)
        (
            fenced_tokens,
            leaf_token_whitespace,
        ) = FencedLeafBlockProcessor.parse_fenced_code_block(
            parser_state,
            position_marker,
            leaf_token_whitespace,
            original_line,
            detabified_original_start_index,
            block_quote_data,
            grab_bag,
        )
        POGGER.debug(">>leaf_token_whitespace>:$:<", leaf_token_whitespace)
        if fenced_tokens:
            new_tokens.extend(fenced_tokens)
            POGGER.debug(">>new_tokens>>$", new_tokens)
        elif parser_state.token_stack[-1].is_fenced_code_block:
            POGGER.debug(">>still in fenced block>:$:<", original_line)
            POGGER.debug(">>leaf_token_whitespace>:$:<", leaf_token_whitespace)
            token_text = position_marker.text_to_parse[position_marker.index_number :]
            if ParserHelper.tab_character in original_line:
                (
                    leaf_token_whitespace,
                    token_text,
                ) = FencedLeafBlockProcessor.__handle_fenced_code_block_with_tab(
                    parser_state,
                    position_marker,
                    original_line,
                    leaf_token_whitespace,
                    token_text,
                )
            new_tokens.append(
                TextMarkdownToken(
                    token_text,
                    leaf_token_whitespace,
                    position_marker=position_marker,
                )
            )
            POGGER.debug(">>new_tokens>>$", new_tokens)
        else:
            return False
        return True

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_fenced_code_block_with_tab_and_extracted_whitespace(
        new_extracted_whitespace: str,
        adj_original_index: int,
        whitespace_start_count: int,
        extracted_whitespace: str,
    ) -> str:
        new_index = 0
        POGGER.debug("new_index>:$:<", new_index)
        while new_index < len(new_extracted_whitespace):
            before_count = new_extracted_whitespace[: new_index + 1]
            POGGER.debug("before_count>:$:<", before_count)
            detabified_before_count = TabHelper.detabify_string(
                before_count, adj_original_index
            )
            POGGER.debug("detabified_before_count>:$:<", detabified_before_count)
            detabified_before_count_length = len(detabified_before_count)
            POGGER.debug(
                "detabified_before_count_length>:$:<",
                detabified_before_count_length,
            )
            POGGER.debug("whitespace_start_count>:$:<", whitespace_start_count)
            if detabified_before_count_length >= whitespace_start_count:
                after_count = new_extracted_whitespace[len(before_count) :]
                break
            new_index += 1
            POGGER.debug("new_index>:$:<", new_index)

        POGGER.debug("new_extracted_whitespace>:$:<", new_extracted_whitespace)
        POGGER.debug(
            "detabified_before_count_length>:$:<", detabified_before_count_length
        )
        POGGER.debug("whitespace_start_count>:$:<", whitespace_start_count)
        if detabified_before_count_length < whitespace_start_count:
            assert (
                before_count == new_extracted_whitespace
            ), "If we do not have enough length, before_count must equal new_extracted_whitespace."
            after_count = ""
        POGGER.debug("before_count>:$:<", before_count)
        POGGER.debug("after_count>:$:<", after_count)

        POGGER.debug("new_extracted_whitespace>:$:<", new_extracted_whitespace)
        POGGER.debug("extracted_whitespace>:$:<", extracted_whitespace)
        replacement_string = after_count or ParserHelper.replace_noop_character

        new_extracted_whitespace = ParserHelper.create_replacement_markers(
            extracted_whitespace, replacement_string
        )
        POGGER.debug("new_extracted_whitespace>:$:<", new_extracted_whitespace)
        return new_extracted_whitespace

    @staticmethod
    def __handle_fenced_code_block_with_tab_starts_tab(
        original_line: str, reconstructed_line: str, resolved_leaf_token_whitespace: str
    ) -> Tuple[str, int, bool]:
        adj_original = reconstructed_line
        adj_original_index = original_line.find(reconstructed_line)
        assert adj_original_index != -1, "reconstructed_line must be in original line."
        assert original_line.endswith(
            reconstructed_line
        ), "Reconstructed_line must end the original line to verify it was built properly."
        original_line_prefix = original_line[: -len(reconstructed_line)]
        split_tab = original_line_prefix.endswith(">")

        resolved_leaf_token_whitespace = TabHelper.detabify_string(
            resolved_leaf_token_whitespace, adj_original_index
        )
        return adj_original, adj_original_index, split_tab

    @staticmethod
    def __handle_fenced_code_block_with_tab_not_starts_tab(
        original_line: str,
        reconstructed_line: str,
        position_marker: PositionMarker,
        was_indented: bool,
    ) -> Tuple[str, int, bool, Optional[str]]:
        if is_bq_start := original_line.startswith(">"):
            indent_prefix = None
        else:
            indent_prefix = " " * position_marker.index_indent
        (
            adj_original,
            adj_original_index,
            split_tab,
        ) = TabHelper.find_tabified_string(
            original_line,
            reconstructed_line,
            abc=is_bq_start,
            was_indented=was_indented,
            reconstruct_prefix=indent_prefix,
        )
        split_tab_whitespace: Optional[str] = None
        if split_tab:
            split_tab_whitespace = original_line[:adj_original_index]
            if not is_bq_start:
                adj_original_index += position_marker.index_indent
        return (
            adj_original,
            adj_original_index,
            split_tab,
            split_tab_whitespace,
        )

    @staticmethod
    def __handle_fenced_code_block_with_tab(
        parser_state: ParserState,
        position_marker: PositionMarker,
        original_line: str,
        leaf_token_whitespace: str,
        token_text: str,
    ) -> Tuple[str, str]:
        fenced_stack_token = cast(
            FencedCodeBlockStackToken, parser_state.token_stack[-1]
        )
        whitespace_start_count = fenced_stack_token.whitespace_start_count
        was_indented = not parser_state.token_stack[-2].is_document

        resolved_leaf_token_whitespace = ParserHelper.remove_all_from_text(
            leaf_token_whitespace
        )
        reconstructed_line = resolved_leaf_token_whitespace + token_text
        split_tab_whitespace: Optional[str] = None
        if reconstructed_line[0] == "\t":
            (
                adj_original,
                adj_original_index,
                split_tab,
            ) = FencedLeafBlockProcessor.__handle_fenced_code_block_with_tab_starts_tab(
                original_line, reconstructed_line, resolved_leaf_token_whitespace
            )
            reconstructed_line_has_tab = split_tab
        else:
            (
                adj_original,
                adj_original_index,
                split_tab,
                split_tab_whitespace,
            ) = FencedLeafBlockProcessor.__handle_fenced_code_block_with_tab_not_starts_tab(
                original_line, reconstructed_line, position_marker, was_indented
            )
            reconstructed_line_has_tab = True

        (
            leaf_token_whitespace,
            token_text,
        ) = FencedLeafBlockProcessor.__handle_fenced_code_block_with_tab_whitespace(
            parser_state,
            leaf_token_whitespace,
            adj_original,
            adj_original_index,
            whitespace_start_count,
            reconstructed_line_has_tab,
            split_tab,
            split_tab_whitespace,
        )
        return leaf_token_whitespace, token_text

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_fenced_code_block_with_tab_whitespace(
        parser_state: ParserState,
        leaf_token_whitespace: str,
        adj_original: str,
        adj_original_index: int,
        whitespace_start_count: int,
        reconstructed_line_has_tab: bool,
        split_tab: bool,
        split_tab_whitespace: Optional[str],
    ) -> Tuple[str, str]:
        space_end_index, extracted_whitespace = ParserHelper.extract_spaces_verified(
            adj_original, 0
        )
        # assert space_end_index != -1

        # detabified_extracted_whitespace = TabHelper.detabify_string(
        #     extracted_whitespace, adj_original_index
        # )
        # assert detabified_extracted_whitespace == resolved_leaf_token_whitespace

        new_extracted_whitespace = extracted_whitespace
        if new_extracted_whitespace and whitespace_start_count:
            # assert not was_indented, "TOoDO: huh?"
            new_extracted_whitespace = FencedLeafBlockProcessor.__handle_fenced_code_block_with_tab_and_extracted_whitespace(
                new_extracted_whitespace,
                adj_original_index,
                whitespace_start_count,
                extracted_whitespace,
            )

        if reconstructed_line_has_tab:
            leaf_token_whitespace = new_extracted_whitespace
        token_text = adj_original[space_end_index:]
        if split_tab:
            TabHelper.adjust_block_quote_indent_for_tab(
                parser_state, split_tab_whitespace
            )

        return leaf_token_whitespace, token_text

    # pylint: enable=too-many-arguments



"""
Module to provide structure to scan through a file.
"""

from abc import ABC, abstractmethod
from typing import List, Optional, Union, cast

from application_properties import ApplicationPropertiesFacade

from pymarkdown.my_application_properties_facade import MyApplicationPropertiesFacade
from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV2,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.setext_heading_markdown_token import SetextHeadingMarkdownToken


class RulePlugin(ABC):
    """
    Class to provide structure to scan through a file.
    Based off of concepts from
    https://github.com/hiddenillusion/example-code/commit/3e2daada652fe9b487574c784e0924bd5fcfe667
    """

    def __init__(self) -> None:
        (
            self.__is_next_token_implemented_in_plugin,
            self.__is_next_line_implemented_in_plugin,
            self.__is_starting_new_file_implemented_in_plugin,
            self.__is_completed_file_implemented_in_plugin,
            self.__is_query_config_implemented_in_plugin,
        ) = (True, True, True, True, False)
        self.__plugin_specific_facade: Optional[MyApplicationPropertiesFacade] = None

    @abstractmethod
    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """

    @property
    def plugin_configuration(self) -> MyApplicationPropertiesFacade:
        """
        Get the configuration facade associated with this plugin.
        """
        assert self.__plugin_specific_facade
        return self.__plugin_specific_facade

    def set_configuration_map(
        self, plugin_specific_facade: ApplicationPropertiesFacade
    ) -> None:
        """
        Set the configuration map with values for the plugin.
        """
        self.__plugin_specific_facade = MyApplicationPropertiesFacade(
            plugin_specific_facade
        )

        self.__is_query_config_implemented_in_plugin = (
            "query_config" in self.__class__.__dict__
        )
        self.__is_next_token_implemented_in_plugin = (
            "next_token" in self.__class__.__dict__
        )
        self.__is_next_line_implemented_in_plugin = (
            "next_line" in self.__class__.__dict__
        )
        self.__is_starting_new_file_implemented_in_plugin = (
            "starting_new_file" in self.__class__.__dict__
        )
        self.__is_completed_file_implemented_in_plugin = (
            "completed_file" in self.__class__.__dict__
        )

    @property
    def is_query_config_implemented_in_plugin(self) -> bool:
        """
        Return whether the query_config function is implemented in the plugin.
        """
        return self.__is_query_config_implemented_in_plugin

    @property
    def is_starting_new_file_implemented_in_plugin(self) -> bool:
        """
        Return whether the starting_new_file function is implemented in the plugin.
        """
        return self.__is_starting_new_file_implemented_in_plugin

    @property
    def is_next_line_implemented_in_plugin(self) -> bool:
        """
        Return whether the next_line function is implemented in the plugin.
        """
        return self.__is_next_line_implemented_in_plugin

    @property
    def is_next_token_implemented_in_plugin(self) -> bool:
        """
        Return whether the next_token function is implemented in the plugin.
        """
        return self.__is_next_token_implemented_in_plugin

    @property
    def is_completed_file_implemented_in_plugin(self) -> bool:
        """
        Return whether the completed_file function is implemented in the plugin.
        """
        return self.__is_completed_file_implemented_in_plugin

    # pylint: disable=too-many-arguments
    def register_fix_token_request(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        plugin_action: str,
        field_name: str,
        field_value: Union[str, int],
    ) -> None:
        """
        Register a request to fix a token.
        """
        context.register_fix_token_request(
            token, self.get_details().plugin_id, plugin_action, field_name, field_value
        )

    def register_replace_tokens_request(
        self,
        context: PluginScanContext,
        start_token: MarkdownToken,
        end_token: MarkdownToken,
        replacement_tokens: List[MarkdownToken],
    ) -> None:
        """
        Register a request to replace a sequence of tokens.
        """
        context.register_replace_tokens_request(
            self.get_details().plugin_id, start_token, end_token, replacement_tokens
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def report_next_line_error(
        self,
        context: PluginScanContext,
        column_number: int,
        line_number_delta: int = 0,
        extra_error_information: Optional[str] = None,
        override_is_error_token_prefaced_by_blank_line: Optional[bool] = None,
    ) -> None:
        """
        Report an error with the current line being processed.
        """
        does_support_fix = False
        plugin_details = self.get_details()
        if isinstance(plugin_details, PluginDetailsV2):
            does_support_fix = plugin_details.plugin_supports_fix

        context.add_triggered_rule(
            context.scan_file,
            context.line_number + line_number_delta,
            column_number,
            plugin_details.plugin_id,
            plugin_details.plugin_name,
            plugin_details.plugin_description,
            extra_error_information,
            does_support_fix,
            override_is_error_token_prefaced_by_blank_line=override_is_error_token_prefaced_by_blank_line,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def report_next_token_error(
        self,
        context: PluginScanContext,
        token: MarkdownToken,
        extra_error_information: Optional[str] = None,
        line_number_delta: int = 0,
        column_number_delta: int = 0,
        use_original_position: bool = False,
        override_is_error_token_prefaced_by_blank_line: Optional[bool] = None,
    ) -> None:
        """
        Report an error with the current token being processed.
        """
        if use_original_position:
            leaf_token = cast(SetextHeadingMarkdownToken, token)
            line_number = leaf_token.original_line_number
            column_number = leaf_token.original_column_number
        else:
            line_number = token.line_number
            column_number = token.column_number
        error_token = token

        does_support_fix = False
        plugin_details = self.get_details()
        # if isinstance(xx, PluginDetailsV2):
        #     xy = cast(PluginDetailsV2, xx)
        #     does_support_fix = xy.plugin_supports_fix

        context.add_triggered_rule(
            context.scan_file,
            line_number + line_number_delta,
            (
                column_number + column_number_delta
                if column_number_delta >= 0
                else -column_number_delta
            ),
            plugin_details.plugin_id,
            plugin_details.plugin_name,
            plugin_details.plugin_description,
            extra_error_information,
            does_support_fix,
            error_token=error_token,
            override_is_error_token_prefaced_by_blank_line=override_is_error_token_prefaced_by_blank_line,
        )

    # pylint: enable=too-many-arguments

    def initialize_from_config(self) -> None:  # noqa: B027
        """
        Event to allow the plugin to load configuration information.
        """

    def query_config(self) -> List[QueryConfigItem]:  # noqa: B027
        """
        Query to find out the configuration that the rule is using.
        """
        return []  # pragma: no cover

    def starting_new_file(self) -> None:  # noqa: B027
        """
        Event that the a new file to be scanned is starting.
        """

    def completed_file(self, context: PluginScanContext) -> None:  # noqa: B027
        """
        Event that the file being currently scanned is now completed.
        """

    def next_line(self, context: PluginScanContext, line: str) -> None:  # noqa: B027
        """
        Event that a new line is being processed.
        """

    def next_token(  # noqa: B027
        self, context: PluginScanContext, token: MarkdownToken
    ) -> None:
        """
        Event that a new token is being processed.
        """



"""
Module to implement a plugin to ensure all files end with a blank line.
"""

from typing import Optional

from pymarkdown.plugin_manager.plugin_details import PluginDetails, PluginDetailsV2
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin


class RuleMd047(RulePlugin):
    """
    Class to implement a plugin to ensure all files end with a blank line.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__last_line: Optional[str] = None
        self.__previous_last_line: Optional[str] = None

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV2(
            plugin_name="single-trailing-newline",
            plugin_id="MD047",
            plugin_enabled_by_default=True,
            plugin_description="Each file should end with a single newline character.",
            plugin_version="0.5.2",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md047.md",
            plugin_supports_fix=True,
            plugin_fix_level=0,
        )

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__last_line = None
        self.__previous_last_line = None

    def next_line(self, context: PluginScanContext, line: str) -> None:
        """
        Event that a new line is being processed.
        """
        _ = context
        if self.__last_line is not None:
            self.__previous_last_line = self.__last_line
        self.__last_line = line

    def completed_file(self, context: PluginScanContext) -> None:
        """
        Event that the file being currently scanned is now completed.
        """
        if context.in_fix_mode:
            if (
                context.last_line_fixed is not None
                and not context.last_line_fixed.endswith("\n")
            ):
                context.set_current_fix_line("\n")
        elif self.__last_line:
            override_is_error_token_prefaced_by_blank_line = (
                self.__previous_last_line is not None
                and not self.__previous_last_line.strip()
            )
            self.report_next_line_error(
                context,
                len(self.__last_line),
                -1,
                override_is_error_token_prefaced_by_blank_line=override_is_error_token_prefaced_by_blank_line,
            )



"""
Module to provide for an encapsulation of the inline uri autolink element.
"""

from typing import Optional, cast

from pymarkdown.inline.inline_helper import InlineHelper
from pymarkdown.tokens.inline_markdown_token import InlineMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class UriAutolinkMarkdownToken(InlineMarkdownToken):
    """
    Class to provide for an encapsulation of the inline uri autolink element.
    """

    __uri_autolink_html_character_escape_map = {
        "<": "&lt;",
        ">": "&gt;",
        "&": "&amp;",
    }

    __raw_html_percent_escape_ascii_chars = '"%[\\]^`{}|'

    # pylint: disable=too-many-arguments
    def __init__(
        self,
        autolink_text: str,
        line_number: int,
        column_number: int,
        add_http_prefix: bool = False,
        add_angle_brackets: bool = True,
    ) -> None:
        self.__autolink_text = autolink_text
        self.__add_http_prefix = add_http_prefix
        self.__add_angle_brackets = add_angle_brackets
        InlineMarkdownToken.__init__(
            self,
            MarkdownToken._token_inline_uri_autolink,
            (
                f"1:{self.__autolink_text}"
                if self.__add_http_prefix
                else self.__autolink_text
            ),
            line_number=line_number,
            column_number=column_number,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_inline_uri_autolink

    # pylint: enable=protected-access

    @property
    def autolink_text(self) -> str:
        """
        Returns the text that is the autolink.
        """
        return self.__autolink_text

    @property
    def add_http_prefix(self) -> bool:
        """
        Returns the whether a http prefix needs to be appended.
        """
        return self.__add_http_prefix

    @property
    def add_angle_brackets(self) -> bool:
        """
        Returns the whether rendered markdown needs to add angle brackets.
        """
        return self.__add_angle_brackets

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            UriAutolinkMarkdownToken,
            UriAutolinkMarkdownToken.__rehydrate_inline_uri_autolink,
            None,
        )

    @staticmethod
    def __rehydrate_inline_uri_autolink(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the uri autolink from the token.
        """
        _ = previous_token

        current_uri_token = cast(UriAutolinkMarkdownToken, current_token)
        if current_uri_token.add_http_prefix:
            return f"{current_uri_token.autolink_text}"
        prefix_char, suffix_char = (
            ("<", ">") if current_uri_token.add_angle_brackets else ("", "")
        )
        return (
            ""
            if context.block_stack[-1].is_inline_link
            else f"{prefix_char}{current_uri_token.autolink_text}{suffix_char}"
        )

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            UriAutolinkMarkdownToken,
            UriAutolinkMarkdownToken.__handle_uri_autolink,
            None,
        )

    @classmethod
    def __handle_uri_autolink(
        cls,
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = transform_state

        autolink_token = cast(UriAutolinkMarkdownToken, next_token)
        in_tag_pretext = InlineHelper.append_text(
            "",
            autolink_token.autolink_text,
            alternate_escape_map=UriAutolinkMarkdownToken.__uri_autolink_html_character_escape_map,
            add_text_signature=False,
        )

        tag_text_parts = []
        for next_character in in_tag_pretext:
            if (
                next_character
                in UriAutolinkMarkdownToken.__raw_html_percent_escape_ascii_chars
            ):
                tag_text_parts.extend(["%", (hex(ord(next_character))[2:]).upper()])
            elif ord(next_character) >= 128:
                encoded_data = next_character.encode("utf8")
                for encoded_byte in encoded_data:
                    tag_text_parts.extend(["%", (hex(encoded_byte)[2:]).upper()])
            else:
                tag_text_parts.append(next_character)

        return "".join(
            [
                output_html,
                '<a href="',
                "http://" if autolink_token.add_http_prefix else "",
                "".join(tag_text_parts),
                '">',
                InlineHelper.append_text(
                    "", autolink_token.autolink_text, add_text_signature=False
                ),
                "</a>",
            ]
        )



"""
Module to implement a plugin that ensures that the style of fenced code blocks is consistent.
"""

from typing import List, cast

from pymarkdown.plugin_manager.plugin_details import (
    PluginDetailsV2,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.fenced_code_block_markdown_token import (
    FencedCodeBlockMarkdownToken,
)
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd048(RulePlugin):
    """
    Class to implement a plugin that ensures that the style of fenced code blocks is consistent.
    """

    __consistent_style = "consistent"
    __tilde_style = "tilde"
    __backtick_style = "backtick"
    __valid_styles = [
        __consistent_style,
        __tilde_style,
        __backtick_style,
    ]

    def __init__(self) -> None:
        super().__init__()
        self.__style_type: str = ""
        self.__actual_style_type: str = ""

    def get_details(self) -> PluginDetailsV2:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="code-fence-style",
            plugin_id="MD048",
            plugin_enabled_by_default=True,
            plugin_description="Code fence style",
            plugin_version="0.6.0",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md048.md",
            plugin_configuration="style",
            plugin_supports_fix=True,
            plugin_fix_level=2,
        )

    @classmethod
    def __validate_configuration_style(cls, found_value: str) -> None:
        if found_value not in RuleMd048.__valid_styles:
            raise ValueError(f"Allowable values: {RuleMd048.__valid_styles}")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__style_type = self.plugin_configuration.get_string_property_with_default(
            "style",
            RuleMd048.__consistent_style,
            valid_value_fn=self.__validate_configuration_style,
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [
            QueryConfigItem("style", self.__style_type),
        ]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__actual_style_type = (
            self.__style_type
            if self.__style_type != RuleMd048.__consistent_style
            else ""
        )

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if not token.is_fenced_code_block:
            return

        fence_token = cast(FencedCodeBlockMarkdownToken, token)
        current_style = (
            RuleMd048.__backtick_style
            if fence_token.fence_character == "`"
            else RuleMd048.__tilde_style
        )
        if not self.__actual_style_type:
            self.__actual_style_type = current_style
        if self.__actual_style_type != current_style:
            if context.in_fix_mode:
                replace_character = (
                    "`"
                    if self.__actual_style_type == RuleMd048.__backtick_style
                    else "~"
                )
                self.register_fix_token_request(
                    context, token, "next_token", "fence_character", replace_character
                )
            else:
                extra_data = (
                    f"Expected: {self.__actual_style_type}; Actual: {current_style}"
                )
                self.report_next_token_error(
                    context, token, extra_error_information=extra_data
                )



from typing import cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_gfm.transform_to_gfm_list_looseness import (
    TransformToGfmListLooseness,
)


class ListStartMarkdownTokenHelper:
    @staticmethod
    def handle_start_list_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        list_token = cast(ListStartMarkdownToken, next_token)
        transform_state.is_in_loose_list = (
            TransformToGfmListLooseness.calculate_list_looseness(
                transform_state.actual_tokens,
                transform_state.actual_token_index,
                list_token,
            )
        )
        if list_token.is_ordered_list_start:
            token_parts = ["<ol"]
            if int(list_token.list_start_content) != 1:
                token_parts.extend(
                    [' start="', str(int(list_token.list_start_content)), '"']
                )
            token_parts.extend([">", ParserHelper.newline_character, "<li>"])
            transform_state.add_leading_text = "".join(token_parts)
        else:
            transform_state.add_leading_text = "".join(
                ["<ul>", ParserHelper.newline_character, "<li>"]
            )
        return output_html

    @staticmethod
    def handle_end_list_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        transform_state.is_in_loose_list = (
            TransformToGfmListLooseness.reset_list_looseness(
                transform_state.actual_tokens,
                transform_state.actual_token_index,
            )
        )
        transform_state.add_trailing_text = "".join(
            [
                "</li>",
                ParserHelper.newline_character,
                "</ul>" if next_token.is_unordered_list_end else "</ol>",
            ]
        )
        return output_html



"""
Module to provide for the creation of tokens for the new lists.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.requeue_line_info import RequeueLineInfo
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.leaf_blocks.leaf_block_processor_paragraph import (
    LeafBlockProcessorParagraph,
)
from pymarkdown.list_blocks.list_block_can_close_helper import ListBlockCanCloseHelper
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.new_list_item_markdown_token import NewListItemMarkdownToken
from pymarkdown.tokens.ordered_list_start_markdown_token import (
    OrderedListStartMarkdownToken,
)
from pymarkdown.tokens.stack_token import (
    BlockQuoteStackToken,
    ListStackToken,
    OrderedListStackToken,
    StackToken,
    UnorderedListStackToken,
)
from pymarkdown.tokens.unordered_list_start_markdown_token import (
    UnorderedListStartMarkdownToken,
)

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-many-lines
# pylint: disable=too-few-public-methods


class ListBlockCreateNewHandler:
    """
    Class to provide for the creation of tokens for the new lists.
    """

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def create_new_list(
        parser_state: ParserState,
        position_marker: PositionMarker,
        indent_level: int,
        extracted_whitespace: str,
        ws_before_marker: int,
        ws_after_marker: int,
        index: int,
        container_level_tokens: List[MarkdownToken],
        remaining_whitespace: int,
        after_marker_ws_index: int,
        current_container_blocks: List[StackToken],
        container_depth: int,
        original_line: str,
        is_ulist: bool,
        alt_adj_ws: Optional[str] = None,
        forced_container_whitespace: Optional[str] = None,
    ) -> Tuple[bool, Optional[str], Optional[RequeueLineInfo]]:
        """
        Create a new list at the current location.
        """
        assert alt_adj_ws is not None
        adj_ws = ListBlockCreateNewHandler.__calculate_create_adj_ws(
            alt_adj_ws, position_marker, parser_state
        )
        found_block_quote_before_list = (
            ListBlockCreateNewHandler.__find_block_quote_before_list(parser_state)
        )
        if found_block_quote_before_list and adj_ws is None and alt_adj_ws is not None:
            adj_ws = alt_adj_ws

        (
            whitespace_to_add,
            alt_adj_ws,
            ws_before_marker,
            indent_level,
        ) = ListBlockCreateNewHandler.__create_new_list_handle_whitespace(
            forced_container_whitespace,
            alt_adj_ws,
            ws_before_marker,
            indent_level,
            extracted_whitespace,
            adj_ws,
        )

        tabbed_whitespace_to_add = None
        tabbed_adjust = -1
        if "\t" in original_line:
            (
                tabbed_whitespace_to_add,
                tabbed_adjust,
            ) = ListBlockCreateNewHandler.__create_new_list_with_tab(
                position_marker, original_line, is_ulist, whitespace_to_add, index
            )

        if is_ulist:
            new_token, new_stack = (
                ListBlockCreateNewHandler.__handle_list_block_unordered(
                    position_marker,
                    indent_level,
                    tabbed_adjust,
                    whitespace_to_add,
                    tabbed_whitespace_to_add,
                    ws_before_marker,
                    ws_after_marker,
                    index,
                )
            )
        else:
            new_token, new_stack = (
                ListBlockCreateNewHandler.__handle_list_block_ordered(
                    position_marker,
                    indent_level,
                    tabbed_adjust,
                    whitespace_to_add,
                    tabbed_whitespace_to_add,
                    ws_before_marker,
                    ws_after_marker,
                    index,
                )
            )
        (
            new_container_level_tokens,
            adjusted_text_to_parse,
            requeue_line_info,
        ) = ListBlockCreateNewHandler.__post_list(
            parser_state,
            new_stack,
            new_token,
            position_marker.text_to_parse,
            remaining_whitespace,
            after_marker_ws_index,
            indent_level,
            current_container_blocks,
            position_marker,
            adj_ws,
            alt_adj_ws,
            container_depth,
        )
        container_level_tokens.extend(new_container_level_tokens)
        return True, adjusted_text_to_parse, requeue_line_info

    # pylint: enable=too-many-arguments, too-many-locals

    @staticmethod
    def __calculate_create_adj_ws(
        adj_ws: str,
        position_marker: PositionMarker,
        parser_state: ParserState,
    ) -> Optional[str]:
        create_adj_ws: Optional[str] = adj_ws
        POGGER.debug("adj_ws=>:$:<", create_adj_ws)
        if position_marker.index_number:
            POGGER.debug("adjusting for nested")
            POGGER.debug("afn>>$", parser_state.token_stack)
            search_index = parser_state.find_last_container_on_stack()
            if parser_state.token_stack[search_index].is_list:
                create_adj_ws = None
        POGGER.debug("create_adj_ws=$=", create_adj_ws)
        return create_adj_ws

    @staticmethod
    def __find_block_quote_before_list(
        parser_state: ParserState,
    ) -> Optional[BlockQuoteStackToken]:
        POGGER.debug_with_visible_whitespace(
            "parser_state.token_stack>$", parser_state.token_stack
        )
        found_block_quote_before_list = None
        token_stack_index = parser_state.find_last_container_on_stack()
        POGGER.debug_with_visible_whitespace("token_stack_index>$", token_stack_index)
        if parser_state.token_stack[token_stack_index].is_list:
            while token_stack_index > 0:
                if parser_state.token_stack[token_stack_index].is_block_quote:
                    found_block_quote_before_list = cast(
                        BlockQuoteStackToken,
                        parser_state.token_stack[token_stack_index],
                    )
                    break
                token_stack_index -= 1
        POGGER.debug_with_visible_whitespace(
            "found_block_quote_before_list>$", found_block_quote_before_list
        )
        return found_block_quote_before_list

    # pylint: disable=too-many-arguments
    @staticmethod
    def __create_new_list_handle_whitespace(
        forced_container_whitespace: Optional[str],
        alt_adj_ws: str,
        ws_before_marker: int,
        indent_level: int,
        extracted_whitespace: str,
        adj_ws: Optional[str],
    ) -> Tuple[str, str, int, int]:
        if forced_container_whitespace is not None:
            whitespace_to_add = (
                forced_container_whitespace + alt_adj_ws
                if alt_adj_ws
                else forced_container_whitespace
            )
            ws_before_marker += len(forced_container_whitespace)
            indent_level += len(forced_container_whitespace)
            alt_adj_ws += forced_container_whitespace
        else:
            whitespace_to_add = extracted_whitespace if adj_ws is None else adj_ws
        return whitespace_to_add, alt_adj_ws, ws_before_marker, indent_level

    # pylint: enable=too-many-arguments
    @staticmethod
    def __create_new_list_with_tab(
        position_marker: PositionMarker,
        original_line: str,
        is_ulist: bool,
        whitespace_to_add: Optional[str],
        index: int,
    ) -> Tuple[Optional[str], int]:
        tabbed_adjust = -1
        (
            tabbed_extract_spaces_index,
            tabbed_extract_spaces,
        ) = ParserHelper.extract_spaces_verified(original_line, 0)
        POGGER.debug("tabbed_extract_spaces_index>:$:<", tabbed_extract_spaces_index)
        POGGER.debug("tabbed_extract_spaces>:$:<", tabbed_extract_spaces)
        POGGER.debug("text_to_parse>:$:<", position_marker.text_to_parse)
        POGGER.debug("index_number>:$:<", position_marker.index_number)
        detabbed_tabbed_extract_spaces = TabHelper.detabify_string(
            tabbed_extract_spaces
        )
        POGGER.debug(
            "detabbed_tabbed_extract_spaces>:$:<", detabbed_tabbed_extract_spaces
        )
        assert (
            detabbed_tabbed_extract_spaces == whitespace_to_add
        ), "two whitespaces must be equal"
        tabbed_whitespace_to_add = (
            tabbed_extract_spaces if "\t" in tabbed_extract_spaces else None
        )

        POGGER.debug("is_ulist>:$:<", is_ulist)

        # parse_index = position_marker.text_to_parse[position_marker.index_number]
        # POGGER.debug("parse_index>:$:<", parse_index)
        parse_index = 1 if is_ulist else index - position_marker.index_number + 1
        POGGER.debug("parse_index>:$:<", parse_index)

        new_index = position_marker.index_number + parse_index
        untabbed_marker = position_marker.text_to_parse[
            position_marker.index_number : new_index
        ]
        POGGER.debug("untabbed_marker>:$:<", untabbed_marker)
        tabbed_marker = original_line[
            tabbed_extract_spaces_index : tabbed_extract_spaces_index + parse_index
        ]
        POGGER.debug("tabbed_marker>:$:<", tabbed_marker)
        assert (
            untabbed_marker == tabbed_marker
        ), "tabbed and untabbed strings must be equal"

        tabbed_extract_spaces_index += len(tabbed_marker)
        POGGER.debug(
            "position_marker.text_to_parse[new_index:]>:$:<",
            position_marker.text_to_parse[new_index:],
        )
        POGGER.debug(
            "original_line[tabbed_extract_spaces_index:]>:$:<",
            original_line[tabbed_extract_spaces_index:],
        )
        if original_line[tabbed_extract_spaces_index] == "\t":
            POGGER.debug("new_index>:$:<", new_index)
            POGGER.debug(
                "tabbed_extract_spaces_index>:$:<", tabbed_extract_spaces_index
            )
            adj_string = TabHelper.detabify_string(
                original_line[tabbed_extract_spaces_index],
                additional_start_delta=new_index,
            )
            POGGER.debug("adj_string>:$:<", adj_string)
            tabbed_adjust = len(adj_string) - 1
            POGGER.debug("tabbed_adjust>:$:<", tabbed_adjust)

        return tabbed_whitespace_to_add, tabbed_adjust

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __post_list(
        parser_state: ParserState,
        new_stack: ListStackToken,
        new_token: ListStartMarkdownToken,
        line_to_parse: str,
        remaining_whitespace: int,
        after_marker_ws_index: int,
        indent_level: int,
        current_container_blocks: List[StackToken],
        position_marker: PositionMarker,
        adj_ws: Optional[str],
        alt_adj_ws: str,
        container_depth: int,
    ) -> Tuple[List[MarkdownToken], Optional[str], Optional[RequeueLineInfo]]:
        """
        Handle the processing of the last part of the list.
        """
        (
            did_find,
            last_list_index,
        ) = LeafBlockProcessorParagraph.check_for_list_in_process(parser_state)
        if did_find:
            POGGER.debug(
                "list-in-process>>$",
                parser_state.token_stack[last_list_index],
            )
            (
                container_level_tokens,
                emit_li,
                requeue_line_info,
            ) = ListBlockCreateNewHandler.__close_required_lists_after_start(
                position_marker,
                parser_state,
                last_list_index,
                new_stack,
                new_token,
                current_container_blocks,
                container_depth,
            )
            if requeue_line_info:
                return [], None, requeue_line_info
        else:
            POGGER.debug(
                "NOT list-in-process>>$",
                parser_state.token_stack[last_list_index],
            )
            container_level_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state, was_forced=True
            )

        assert (
            container_level_tokens is not None
        ), "Posting some token must be the result of one of these actions."
        POGGER.debug("__post_list>>before>>$", container_level_tokens)
        if not did_find or not emit_li:
            POGGER.debug("__post_list>>adding>>$", new_token)
            parser_state.token_stack.append(new_stack)
            container_level_tokens.append(new_token)
        else:
            POGGER.debug("__post_list>>new list item>>")
            assert emit_li, "if here, emitting a new list item, not a start"
            ListBlockCreateNewHandler.__post_list_use_new_list_item(
                parser_state,
                new_token,
                container_level_tokens,
                indent_level,
                position_marker,
                adj_ws,
                alt_adj_ws,
            )
        parser_state.set_no_para_start_if_empty()
        padded_spaces = ParserHelper.repeat_string(
            ParserHelper.space_character, remaining_whitespace
        )
        return (
            container_level_tokens,
            f"{padded_spaces}{line_to_parse[after_marker_ws_index:]}",
            None,
        )
        # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __post_list_use_new_list_item(
        parser_state: ParserState,
        new_token: ListStartMarkdownToken,
        container_level_tokens: List[MarkdownToken],
        indent_level: int,
        position_marker: PositionMarker,
        adj_ws: Optional[str],
        alt_adj_ws: str,
    ) -> None:
        POGGER.debug("instead of-->$", new_token)

        stack_index = len(parser_state.token_stack) - 1
        while stack_index and not parser_state.token_stack[stack_index].is_list:
            stack_index -= 1
        if stack_index != len(parser_state.token_stack) - 1:
            POGGER.debug("stack_index>$", stack_index)
            POGGER.debug("parser_state.token_stack>$", parser_state.token_stack)
            POGGER.debug(
                "len(parser_state.token_stack)>$", len(parser_state.token_stack) - 1
            )
            new_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                until_this_index=stack_index + 1,
                include_block_quotes=True,
            )
            POGGER.debug("new_tokens>$", new_tokens)
            POGGER.debug("parser_state.token_stack>$", parser_state.token_stack)
            POGGER.debug(
                "len(parser_state.token_stack)>$", len(parser_state.token_stack) - 1
            )
            container_level_tokens.extend(new_tokens)

        top_stack_item = parser_state.token_stack[-1]
        assert top_stack_item.is_list, "top stack item must be a list."
        top_stack_list_token = cast(ListStackToken, top_stack_item)
        POGGER.debug("new_token>$", new_token)
        POGGER.debug("top_stack_item>$", top_stack_list_token)
        POGGER.debug(
            "top_stack_item.mmt>$", top_stack_list_token.matching_markdown_token
        )
        list_start_content = (
            new_token.list_start_content if new_token.is_ordered_list_start else ""
        )

        POGGER.debug("adj_ws-->:$:<", adj_ws)
        POGGER.debug("alt_adj_ws-->:$:<", alt_adj_ws)
        exws = alt_adj_ws if adj_ws is None else new_token.extracted_whitespace

        # Replace the "other" list start token with a new list item token.
        # The overwritting of the value of new_token is specifically called for.
        replacement_token = NewListItemMarkdownToken(
            indent_level,
            position_marker,
            exws,
            list_start_content,
        )
        top_stack_list_token.set_last_new_list_token(replacement_token)
        container_level_tokens.append(replacement_token)

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __close_required_lists_after_start(
        position_marker: PositionMarker,
        parser_state: ParserState,
        last_list_index: int,
        new_stack: ListStackToken,
        new_token: ListStartMarkdownToken,
        current_container_blocks: List[StackToken],
        container_depth: int,
    ) -> Tuple[
        Optional[List[MarkdownToken]], Optional[bool], Optional[RequeueLineInfo]
    ]:
        """
        After a list start, check to see if any others need closing.
        """
        POGGER.debug("list-in-process>>$", parser_state.token_stack[last_list_index])
        POGGER.debug(
            "list-in-process.token>>$",
            parser_state.token_stack[last_list_index].matching_markdown_token,
        )
        POGGER.debug("new_token>>$", new_token)
        (
            container_level_tokens,
            requeue_line_info,
        ) = parser_state.close_open_blocks_fn(
            parser_state,
            until_this_index=last_list_index + 1,
            caller_can_handle_requeue=True,
            requeue_reset=True,
        )
        if requeue_line_info:
            return None, None, requeue_line_info

        repeat_check, emit_li_token_instead_of_list_start_token = True, False

        POGGER.debug("token_stack>>$", parser_state.token_stack)
        if (
            not container_depth
            and len(parser_state.token_stack) > 1
            and parser_state.token_stack[1].is_block_quote
        ):
            extra_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                until_this_index=1,
                include_lists=True,
                include_block_quotes=True,
            )
            POGGER.debug("extra_tokens>>$", extra_tokens)
            container_level_tokens.extend(extra_tokens)
            POGGER.debug("token_stack>>$", parser_state.token_stack)
            repeat_check = False

        POGGER.debug("old-stack>>$<<", container_level_tokens)
        while repeat_check:
            (
                repeat_check,
                emit_li_token_instead_of_list_start_token,
                last_list_index,
            ) = ListBlockCreateNewHandler.__close_next_level_of_lists(
                position_marker,
                parser_state,
                new_stack,
                new_token,
                current_container_blocks,
                container_level_tokens,
                last_list_index,
                container_depth,
            )
        POGGER.debug("token_stack>>$", parser_state.token_stack)
        POGGER.debug("container_level_tokens>>$", container_level_tokens)
        POGGER.debug(
            "emit_li_token_instead_of_list_start_token>>$",
            emit_li_token_instead_of_list_start_token,
        )
        return container_level_tokens, emit_li_token_instead_of_list_start_token, None

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __close_next_level_of_lists(
        position_marker: PositionMarker,
        parser_state: ParserState,
        new_stack: ListStackToken,
        new_token: ListStartMarkdownToken,
        current_container_blocks: List[StackToken],
        container_level_tokens: List[MarkdownToken],
        last_list_index: int,
        container_depth: int,
    ) -> Tuple[bool, bool, int]:
        POGGER.debug("start")

        (
            do_not_emit,
            emit_li_token_instead_of_list_start_token,
            extra_tokens,
        ) = ListBlockCreateNewHandler.__are_list_starts_equal(
            position_marker,
            parser_state,
            last_list_index,
            new_stack,
            current_container_blocks,
            container_depth,
        )
        POGGER.debug("extra_tokens>>$", extra_tokens)
        POGGER.debug(
            "emit_li_token_instead_of_list_start_token>>$",
            emit_li_token_instead_of_list_start_token,
        )
        POGGER.debug("do_not_emit>>$", do_not_emit)
        container_level_tokens.extend(extra_tokens)
        repeat_check = False
        if do_not_emit:
            (
                last_list_index,
                repeat_check,
            ) = ListBlockCreateNewHandler.__close_next_level_of_lists_do_not_emit(
                parser_state,
                new_stack,
                new_token,
                current_container_blocks,
                emit_li_token_instead_of_list_start_token,
                container_level_tokens,
                container_depth,
            )
        else:
            POGGER.debug("post_list>>close open blocks and emit")
            (
                repeat_check,
                last_list_index,
            ) = ListBlockCreateNewHandler.__close_next_level_of_lists_do_emit(
                parser_state,
                last_list_index,
                container_level_tokens,
                new_stack,
                new_token,
            )
        return repeat_check, emit_li_token_instead_of_list_start_token, last_list_index

    @staticmethod
    def __close_next_level_of_lists_do_emit(
        parser_state: ParserState,
        last_list_index: int,
        container_level_tokens: List[MarkdownToken],
        new_stack: ListStackToken,
        new_token: ListStartMarkdownToken,
    ) -> Tuple[bool, int]:
        close_tokens, _ = parser_state.close_open_blocks_fn(
            parser_state,
            until_this_index=last_list_index,
            include_lists=True,
            include_block_quotes=True,
        )
        container_level_tokens.extend(close_tokens)

        (
            did_find,
            last_list_index,
        ) = LeafBlockProcessorParagraph.check_for_list_in_process(parser_state)
        POGGER.debug(
            "did_find>>$--last_list_index--$",
            did_find,
            last_list_index,
        )
        repeat_check = False
        if did_find:
            last_list_stack_token = cast(
                ListStackToken, parser_state.token_stack[last_list_index]
            )
            POGGER.debug(
                "ARE-EQUAL>>stack>>$>>new>>$",
                last_list_stack_token,
                new_stack,
            )
            POGGER.debug(
                "ARE-EQUAL>>stack>>$>>new>>$",
                last_list_stack_token.matching_markdown_token,
                new_token,
            )
            last_list_markdown_token = cast(
                ListStartMarkdownToken,
                last_list_stack_token.matching_markdown_token,
            )
            old_indent = (
                len(last_list_markdown_token.list_start_content)
                if last_list_markdown_token.is_ordered_list_start
                else 2
            )
            POGGER.debug(
                "new_token.column_number($) <= old_indent($)",
                new_token.column_number,
                old_indent,
            )
            repeat_check = new_token.column_number <= last_list_stack_token.indent_level
            POGGER.debug(
                "repeat_check($) = new_token.column_number($) - last_list_stack_token.indent_level($)",
                repeat_check,
                new_token.column_number,
                last_list_stack_token.indent_level,
            )
        return repeat_check, last_list_index

    @staticmethod
    def __close_next_level_of_lists_do_not_emit(
        parser_state: ParserState,
        new_stack: ListStackToken,
        new_token: ListStartMarkdownToken,
        current_container_blocks: List[StackToken],
        emit_li_token_instead_of_list_start_token: bool,
        container_level_tokens: List[MarkdownToken],
        container_depth: int,
    ) -> Tuple[int, bool]:
        POGGER.debug("post_list>>don't emit")
        (
            did_find,
            last_list_index,
        ) = LeafBlockProcessorParagraph.verify_list_in_process(parser_state)
        last_list_index_token = cast(
            ListStackToken, parser_state.token_stack[last_list_index]
        )
        POGGER.debug("parser_state.token_stack>>$", parser_state.token_stack)
        POGGER.debug(
            "did_find>>$--last_list_index--$",
            did_find,
            last_list_index,
        )
        POGGER.debug(
            "ARE-EQUAL>>stack>>$>>new>>$",
            last_list_index_token,
            new_stack,
        )
        repeat_check = not (
            last_list_index_token.type_name == new_stack.type_name
            or new_stack.start_index > last_list_index_token.start_index
        )
        POGGER.debug("current_container_blocks>>$", current_container_blocks)
        POGGER.debug(
            "emit_li_token_instead_of_list_start_token>:$:  repeat_check:$:",
            emit_li_token_instead_of_list_start_token,
            repeat_check,
        )

        if not repeat_check and not emit_li_token_instead_of_list_start_token:
            ListBlockCreateNewHandler.__close_next_level_of_lists_do_not_emit_cleanup(
                parser_state,
                last_list_index_token,
                new_token,
                last_list_index,
                container_level_tokens,
                container_depth,
            )
        return last_list_index, repeat_check

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __close_next_level_of_lists_do_not_emit_cleanup(
        parser_state: ParserState,
        last_list_index_token: ListStackToken,
        new_token: ListStartMarkdownToken,
        last_list_index: int,
        container_level_tokens: List[MarkdownToken],
        container_depth: int,
    ) -> None:
        parent_list_indent = last_list_index_token.indent_level
        POGGER.debug("parent_list_indent>>$", parent_list_indent)
        new_token_column_number = new_token.column_number
        POGGER.debug("new_token_column_number>>$", new_token_column_number)
        assert (
            parser_state.original_line_to_parse is not None
        ), "Original line must have been defined by now."
        intermediate_line_content = parser_state.original_line_to_parse[
            parent_list_indent : new_token_column_number - 1
        ]
        POGGER.debug("intermediate_line_content:$:", intermediate_line_content)
        if ">" not in intermediate_line_content:
            close_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                until_this_index=last_list_index,
                include_block_quotes=True,
            )
            if close_tokens:
                container_level_tokens.extend(close_tokens)
                assert (
                    not container_depth
                ), "If here, we must have a non-zero container depth."
                list_token = cast(
                    ListStartMarkdownToken,
                    last_list_index_token.matching_markdown_token,
                )
                delta = new_token.column_number - list_token.column_number
                new_token.set_extracted_whitespace(
                    "".rjust(delta, ParserHelper.space_character)
                )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __are_list_starts_equal(
        position_marker: PositionMarker,
        parser_state: ParserState,
        last_list_index: int,
        new_stack: ListStackToken,
        current_container_blocks: List[StackToken],
        container_depth: int,
    ) -> Tuple[bool, bool, List[MarkdownToken]]:
        """
        Check to see if the list starts are equal, and hence a continuation of
        the current list.
        """

        POGGER.debug(
            "ARE-EQUAL>>stack>>$>>new>>$",
            parser_state.token_stack[last_list_index],
            new_stack,
        )
        if parser_state.token_stack[last_list_index] == new_stack:
            balancing_tokens, _ = parser_state.close_open_blocks_fn(
                parser_state,
                until_this_index=last_list_index,
                include_block_quotes=True,
            )
            return True, True, balancing_tokens

        document_token_index = len(parser_state.token_document) - 1
        while document_token_index >= 0 and not (
            parser_state.token_document[document_token_index].is_any_list_token
        ):
            document_token_index -= 1
        assert document_token_index >= 0, "List token must be found."
        document_list_token = cast(
            ListStartMarkdownToken, parser_state.token_document[document_token_index]
        )
        last_list_stack_token = cast(
            ListStackToken, parser_state.token_stack[last_list_index]
        )

        POGGER.debug("parser_state.token_document=$", parser_state.token_document)
        POGGER.debug("parser_state.token_stack=$", parser_state.token_stack)
        POGGER.debug("ARE-EQUAL>>Last_List_token=$", document_list_token)
        old_start_index, old_last_marker_character, current_start_index = (
            document_list_token.indent_level,
            last_list_stack_token.list_character[-1],
            new_stack.ws_before_marker,
        )
        POGGER.debug(
            "old>>$>>$",
            last_list_stack_token.extra_data,
            old_last_marker_character,
        )
        POGGER.debug("new>>$>>$", new_stack.extra_data, new_stack.list_character[-1])
        return ListBlockCreateNewHandler.__implement_based_on_equality(
            parser_state,
            position_marker,
            last_list_stack_token,
            old_last_marker_character,
            new_stack,
            current_start_index,
            old_start_index,
            current_container_blocks,
            container_depth,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __implement_based_on_equality(
        parser_state: ParserState,
        position_marker: PositionMarker,
        last_list_stack_token: ListStackToken,
        old_last_marker_character: str,
        new_stack: ListStackToken,
        current_start_index: int,
        old_start_index: int,
        current_container_blocks: List[StackToken],
        container_depth: int,
    ) -> Tuple[bool, bool, List[MarkdownToken]]:
        ## Determine if the new start is within the range of the old stack.  If so
        ## AND some combination of the IF statement, then switch.
        ## i.e. a + at col 1 followed by a - at column 1 is a new list
        ## i.e. a + at col 1 followed by a - at column 3 is a new sublist
        last_list_indent = (
            last_list_stack_token.last_new_list_token.indent_level
            if last_list_stack_token.last_new_list_token
            else last_list_stack_token.indent_level
        )
        POGGER.debug(
            "position_marker.index_number>>$ >= xx>>$",
            position_marker,
            last_list_indent,
        )

        is_indented_enough = position_marker.index_number >= last_list_indent
        POGGER.debug("is_indented_enough>>$", is_indented_enough)
        if (
            is_indented_enough
            or old_last_marker_character == new_stack.list_character[-1]
            and last_list_stack_token.type_name == new_stack.type_name
        ):
            balancing_tokens: List[MarkdownToken] = []
            POGGER.debug("new_stack>$<", new_stack)
            POGGER.debug("new_stack>$<", new_stack.matching_markdown_token)
            POGGER.debug("old_stack>$<", last_list_stack_token)
            POGGER.debug(
                "old_stack>$<",
                last_list_stack_token.matching_markdown_token,
            )
            emit_li_token_instead_of_list_start_token = (
                ListBlockCreateNewHandler.__process_eligible_list_start(
                    parser_state,
                    balancing_tokens,
                    current_start_index,
                    old_start_index,
                    current_container_blocks,
                    new_stack,
                    last_list_stack_token,
                )
            )
            return True, emit_li_token_instead_of_list_start_token, balancing_tokens

        POGGER.debug("SUBLIST WITH DIFFERENT")
        POGGER.debug("container_depth:$:", container_depth)
        POGGER.debug("are_list_starts_equal>>ELIGIBLE!!!")
        POGGER.debug(
            "are_list_starts_equal>>current_start_index>>$>>old_start_index>>$",
            current_start_index,
            old_start_index,
        )
        empty_balancing_tokens: List[MarkdownToken] = []
        return current_start_index >= old_start_index, False, empty_balancing_tokens

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __process_eligible_list_start(
        parser_state: ParserState,
        balancing_tokens: List[MarkdownToken],
        current_start_index: int,
        old_start_index: int,
        current_container_blocks: List[StackToken],
        new_stack: StackToken,
        last_list_stack_token: ListStackToken,
    ) -> bool:
        POGGER.debug("are_list_starts_equal>>ELIGIBLE!!!")
        POGGER.debug("current_container_blocks>>$", current_container_blocks)
        POGGER.debug(
            "are_list_starts_equal>>current_start_index>>$>>old_start_index>>$",
            current_start_index,
            old_start_index,
        )
        POGGER.debug("last_list_stack_token>>$", last_list_stack_token)
        last_list_markdown_token = cast(
            ListStartMarkdownToken, last_list_stack_token.matching_markdown_token
        )
        POGGER.debug("last_list_markdown_token>>$", last_list_markdown_token)

        last_list_indent = (
            last_list_stack_token.last_new_list_token.indent_level
            if last_list_stack_token.last_new_list_token
            else last_list_markdown_token.indent_level
        )

        POGGER.debug(
            "current_start_index>>$ >= last_list_indent>>$",
            last_list_stack_token,
            last_list_indent,
        )
        if current_start_index >= last_list_indent:
            return False

        POGGER.debug("current_container_blocks>>$", current_container_blocks)
        if len(current_container_blocks) > 1:
            POGGER.debug("current_container_blocks-->$", parser_state.token_stack)
            allow_list_removal = ListBlockCanCloseHelper.calculate_can_remove_list(
                parser_state, current_start_index
            )
            POGGER.debug("allow_list_removal-->$", allow_list_removal)
            ListBlockCanCloseHelper.close_required_lists(
                parser_state,
                allow_list_removal,
                balancing_tokens,
                new_stack,
            )

        return True

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_list_block_unordered(
        position_marker: PositionMarker,
        indent_level: int,
        tabbed_adjust: int,
        extracted_whitespace: str,
        tabbed_whitespace_to_add: Optional[str],
        ws_before_marker: int,
        ws_after_marker: int,
        index: int,
    ) -> Tuple[ListStartMarkdownToken, ListStackToken]:
        # This is done to allow for this function and __handle_list_block_ordered
        # to be called using the same pattern.
        _ = index

        new_token = UnorderedListStartMarkdownToken(
            position_marker.text_to_parse[position_marker.index_number],
            indent_level,
            tabbed_adjust,
            extracted_whitespace,
            tabbed_whitespace_to_add,
            position_marker,
        )

        POGGER.debug("unordered-token-->$", new_token)

        return new_token, UnorderedListStackToken(
            indent_level,
            position_marker.text_to_parse[position_marker.index_number],
            ws_before_marker,
            ws_after_marker,
            position_marker.index_number,
            new_token,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    @staticmethod
    def __handle_list_block_ordered(
        position_marker: PositionMarker,
        indent_level: int,
        tabbed_adjust: int,
        extracted_whitespace: str,
        tabbed_whitespace_to_add: Optional[str],
        ws_before_marker: int,
        ws_after_marker: int,
        index: int,
    ) -> Tuple[ListStartMarkdownToken, ListStackToken]:
        new_token = OrderedListStartMarkdownToken(
            position_marker.text_to_parse[index],
            position_marker.text_to_parse[position_marker.index_number : index],
            indent_level,
            tabbed_adjust,
            extracted_whitespace,
            tabbed_whitespace_to_add,
            position_marker,
        )

        POGGER.debug("ordered-token-->$", new_token)

        return new_token, OrderedListStackToken(
            indent_level,
            position_marker.text_to_parse[position_marker.index_number : index + 1],
            ws_before_marker,
            ws_after_marker,
            position_marker.index_number,
            new_token,
        )

    # pylint: enable=too-many-arguments


# pylint: enable=too-few-public-methods



"""
Module to provide processing for non-fenced-code-block sections.
"""

import logging
from typing import List, Optional, Tuple, cast

from pymarkdown.block_quotes.block_quote_count_helper import BlockQuoteCountHelper
from pymarkdown.block_quotes.block_quote_data import BlockQuoteData
from pymarkdown.general.constants import Constants
from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.parser_state import ParserState
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.general.requeue_line_info import RequeueLineInfo
from pymarkdown.general.tab_helper import TabHelper
from pymarkdown.tokens.block_quote_markdown_token import BlockQuoteMarkdownToken
from pymarkdown.tokens.list_start_markdown_token import ListStartMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.stack_token import BlockQuoteStackToken, ListStackToken

POGGER = ParserLogger(logging.getLogger(__name__))

# pylint: disable=too-few-public-methods


class BlockQuoteNonFencedHelper:
    """
    Class to provide processing for non-fenced-code-block sections.
    """

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def handle_non_fenced_code_section(
        parser_state: ParserState,
        block_quote_data: BlockQuoteData,
        extracted_whitespace: str,
        position_marker: PositionMarker,
        original_start_index: int,
        container_start_bq_count: int,
        line_to_parse: str,
        start_index: int,
        leaf_tokens: List[MarkdownToken],
        original_line: str,
        last_block_quote_index: int,
        avoid_block_starts: bool,
    ) -> Tuple[
        str,
        int,
        List[MarkdownToken],
        List[MarkdownToken],
        BlockQuoteData,
        int,
        bool,
        int,
        Optional[str],
        bool,
        Optional[RequeueLineInfo],
        bool,
    ]:
        """
        Handle a non-fenced code block within a block quote.
        """

        POGGER.debug("handle_block_quote_section>>not fenced")
        (
            container_level_tokens,
            requeue_line_info,
            force_list_continuation,
            block_quote_data,
        ) = BlockQuoteCountHelper.ensure_stack_at_level(
            parser_state,
            block_quote_data,
            extracted_whitespace,
            position_marker,
            original_start_index,
            container_start_bq_count,
        )
        POGGER.debug("force_list_continuation=$", force_list_continuation)
        if requeue_line_info:
            return (
                line_to_parse,
                start_index,
                leaf_tokens,
                container_level_tokens,
                block_quote_data,
                0,
                False,
                last_block_quote_index,
                None,
                avoid_block_starts,
                requeue_line_info,
                force_list_continuation,
            )
        # POGGER.debug("extracted_whitespace:$:", extracted_whitespace)
        # POGGER.debug("line_to_parse:$:", line_to_parse)
        # POGGER.debug("start_index:$:", start_index)
        # POGGER.debug(
        #     "position_marker.index_number:$:", position_marker.index_number
        # )
        # POGGER.debug(
        #     "position_marker.index_indent:$:", position_marker.index_indent
        # )
        removed_text = f"{extracted_whitespace}{line_to_parse[position_marker.index_number : start_index]}"
        # POGGER.debug(
        #     "==EWS[$],OSI[$],SI[$],LTP[$],RT=[$]",
        #     extracted_whitespace,
        #     original_start_index,
        #     position_marker.index_number,
        #     position_marker.text_to_parse,
        #     removed_text,
        # )

        (
            line_to_parse,
            removed_chars_at_start,
            text_removed_by_container,
            did_blank,
            leaf_tokens,
            requeue_line_info,
        ) = BlockQuoteNonFencedHelper.__handle_non_fenced_code_section_no_requeue(
            parser_state,
            position_marker,
            line_to_parse,
            start_index,
            removed_text,
            container_start_bq_count,
            block_quote_data,
            original_start_index,
            # extra_consumed_whitespace,
            container_level_tokens,
            original_line,
        )
        return (
            line_to_parse,
            start_index,
            leaf_tokens,
            container_level_tokens,
            block_quote_data,
            removed_chars_at_start,
            did_blank,
            last_block_quote_index,
            text_removed_by_container,
            avoid_block_starts,
            requeue_line_info,
            force_list_continuation,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments, too-many-locals
    @staticmethod
    def __handle_non_fenced_code_section_no_requeue(
        parser_state: ParserState,
        position_marker: PositionMarker,
        line_to_parse: str,
        start_index: int,
        removed_text: str,
        container_start_bq_count: int,
        block_quote_data: BlockQuoteData,
        original_start_index: int,
        # extra_consumed_whitespace: Optional[int],
        container_level_tokens: List[MarkdownToken],
        original_line: str,
    ) -> Tuple[str, int, str, bool, List[MarkdownToken], Optional[RequeueLineInfo]]:
        (
            line_to_parse,
            stack_index,
        ) = (
            line_to_parse[start_index:],
            parser_state.find_last_block_quote_on_stack(),
        )
        POGGER.debug("==REM[$],LTP[$]", removed_text, line_to_parse)

        assert (
            stack_index != -1
        ), "There must be a block quote on the stack if we are in here."
        found_bq_stack_token = cast(
            BlockQuoteStackToken, parser_state.token_stack[stack_index]
        )
        is_not_blank_line = bool(line_to_parse.strip(Constants.ascii_whitespace))

        BlockQuoteNonFencedHelper.__do_block_quote_leading_spaces_adjustments(
            parser_state,
            stack_index,
            container_start_bq_count,
            block_quote_data,
            removed_text,
            found_bq_stack_token,
            removed_text,
            original_start_index,
            # extra_consumed_whitespace,
            container_level_tokens,
            original_line,
            is_not_blank_line,
            position_marker,
        )
        POGGER.debug("text_removed_by_container=[$]", removed_text)
        POGGER.debug("removed_text=[$]", removed_text)
        if is_not_blank_line:
            return (line_to_parse, start_index, removed_text, False, [], None)
        did_blank, leaf_tokens, requeue_line_info = (
            BlockQuoteNonFencedHelper.__handle_normal_blank_line(
                parser_state,
                block_quote_data,
                position_marker,
                removed_text,
                line_to_parse,
            )
        )
        if requeue_line_info:
            container_level_tokens.extend(leaf_tokens)
            leaf_tokens = []
        return (
            line_to_parse,
            start_index,
            removed_text,
            did_blank,
            leaf_tokens,
            requeue_line_info,
        )

    # pylint: enable=too-many-arguments, too-many-locals

    # pylint: disable=too-many-arguments
    @staticmethod
    def __do_block_quote_leading_spaces_adjustments(
        parser_state: ParserState,
        stack_index: int,
        container_start_bq_count: int,
        block_quote_data: BlockQuoteData,
        text_removed_by_container: str,
        found_bq_stack_token: BlockQuoteStackToken,
        removed_text: str,
        original_start_index: int,
        container_level_tokens: List[MarkdownToken],
        original_line: str,
        is_not_blank_line: bool,
        position_marker: PositionMarker,
    ) -> None:
        POGGER.debug("__hbqs>>removed_text>>:$:<", removed_text)
        POGGER.debug("__hbqs>>container_start_bq_count>>$", container_start_bq_count)
        POGGER.debug("__hbqs>>original_start_index>>$", original_start_index)
        POGGER.debug("token_stack--$", parser_state.token_stack)
        original_start_index = BlockQuoteNonFencedHelper.__block_quote_start_adjust(
            parser_state, original_start_index, container_level_tokens
        )
        # original_removed_text = removed_text
        adjusted_removed_text = (
            removed_text[original_start_index:]
            if container_start_bq_count and original_start_index
            else removed_text
        )

        POGGER.debug("dbqlsa>>adjusted_removed_text>>:$:<", adjusted_removed_text)
        (
            adjusted_removed_text,
            tabbed_removed_text,
        ) = BlockQuoteNonFencedHelper.__adjust_1(
            parser_state,
            container_start_bq_count,
            adjusted_removed_text,
            text_removed_by_container,
            stack_index,
            block_quote_data,
            original_line,
        )
        POGGER.debug("dbqlsa>>adjusted_removed_text>>:$:<", adjusted_removed_text)
        POGGER.debug("dbqlsa>>tabbed_removed_text>>:$:<", tabbed_removed_text)

        POGGER.debug("__hbqs>>adjusted_removed_text>>:$:<", adjusted_removed_text)
        POGGER.debug("__hbqs>>tabbed_removed_text>>:$:<", tabbed_removed_text)
        POGGER.debug("token_stack--$", parser_state.token_stack)
        POGGER.debug("token_document--$", parser_state.token_document)
        POGGER.debug("dbqlsa>>found_bq_stack_token>>$", found_bq_stack_token)

        POGGER.debug("dbqlsa>>adjusted_removed_text>>:$:<<", adjusted_removed_text)
        special_case = False
        POGGER.debug("dbqlsa>>adjusted_removed_text>>:$:<<", adjusted_removed_text)
        POGGER.debug("dbqlsa>>special_case>>$", special_case)

        BlockQuoteNonFencedHelper.__do_block_quote_leading_spaces_adjustments_adjust_bleading(
            parser_state,
            found_bq_stack_token,
            tabbed_removed_text,
            adjusted_removed_text,
            special_case,
            is_not_blank_line,
            stack_index,
            position_marker,
            text_removed_by_container,
        )

    # pylint: enable=too-many-arguments

    @staticmethod
    def __handle_normal_blank_line(
        parser_state: ParserState,
        block_quote_data: BlockQuoteData,
        position_marker: PositionMarker,
        text_removed_by_container: str,
        line_to_parse: str,
    ) -> Tuple[bool, List[MarkdownToken], Optional[RequeueLineInfo]]:
        POGGER.debug("call __handle_block_quote_section>>handle_blank_line")

        POGGER.debug(
            "__hbqs>>block_quote_data.current_count>>$", block_quote_data.current_count
        )
        POGGER.debug("__hbqs>>token_stack>>$", parser_state.token_stack)

        adjusted_position_marker = PositionMarker(
            position_marker.line_number,
            len(text_removed_by_container),
            position_marker.text_to_parse,
        )
        (leaf_tokens, requeue_line_info) = parser_state.handle_blank_line_fn(
            parser_state,
            line_to_parse,
            from_main_transform=False,
            position_marker=adjusted_position_marker,
        )
        assert (
            leaf_tokens is not None
        ), "Handling of blank lines always produces tokens."
        POGGER.debug("handle_block_quote_section>>leaf_tokens>>$", leaf_tokens)
        # assert not (
        #     requeue_line_info and requeue_line_info.lines_to_requeue
        # ), "No handling of requeuing available here."

        return True, leaf_tokens, requeue_line_info

    # pylint: disable=too-many-arguments
    @staticmethod
    def __adjust_1(
        parser_state: ParserState,
        container_start_bq_count: int,
        adjusted_removed_text: str,
        text_removed_by_container: str,
        stack_index: int,
        block_quote_data: BlockQuoteData,
        original_line: str,
    ) -> Tuple[str, Optional[str]]:
        POGGER.debug("__hbqs>>container_start_bq_count>>$", container_start_bq_count)
        POGGER.debug(
            "__hbqs>>token_stack[stack_index - 1]>>$",
            parser_state.token_stack[stack_index - 1],
        )
        if (
            container_start_bq_count
            and parser_state.token_stack[stack_index - 1].is_block_quote
        ):
            block_stack_token = cast(
                BlockQuoteStackToken, parser_state.token_stack[stack_index - 1]
            )
            count_of_actual_starts = ParserHelper.count_characters_in_text(
                adjusted_removed_text, ">"
            )
            assert (
                count_of_actual_starts != block_quote_data.current_count
            ), "Block quote start counts are expected to be different"
            block_quote_token = cast(
                BlockQuoteMarkdownToken, block_stack_token.matching_markdown_token
            )
            adj_leading_spaces = block_quote_token.bleading_spaces
            assert adj_leading_spaces is not None, "Leading spaces should not be None."
            POGGER.debug("__hbqs>>count_of_actual_starts>>$", count_of_actual_starts)
            POGGER.debug("__hbqs>>original_line>:$:<", original_line)
            POGGER.debug(
                "__hbqs>>adj_leading_spaces>>:$:$:<",
                len(adj_leading_spaces),
                adj_leading_spaces,
            )
            POGGER.debug(
                "__hbqs>>adjusted_removed_text>>:$:$:<",
                len(adjusted_removed_text),
                adjusted_removed_text,
            )
            POGGER.debug(
                "__hbqs>>text_removed_by_container>>:$:$:<",
                len(text_removed_by_container),
                text_removed_by_container,
            )
            last_line_index = adj_leading_spaces.rfind("\n")
            if last_line_index != -1:
                adj_leading_spaces = adj_leading_spaces[last_line_index + 1 :]

            delta = len(text_removed_by_container) - len(
                adj_leading_spaces + adjusted_removed_text
            )
            adj_leading_spaces = adj_leading_spaces + ParserHelper.repeat_string(
                ParserHelper.space_character, delta
            )
            adjusted_removed_text = adj_leading_spaces + adjusted_removed_text

            POGGER.debug("__hbqs>>adjusted_removed_text>>:$:<", adjusted_removed_text)

        POGGER.debug("__hbqs>>adjusted_removed_text>>:$:<", adjusted_removed_text)
        adjusted_removed_text, tabbed_removed_text = (
            BlockQuoteNonFencedHelper.__adjust_1_with_tab(
                original_line, adjusted_removed_text
            )
        )
        return (adjusted_removed_text, tabbed_removed_text)

    # pylint: enable=too-many-arguments

    @staticmethod
    def __adjust_1_with_tab_part_1(
        detabified_original_line_prefix: str, adjusted_removed_text: str
    ) -> str:
        _, dolp_start_whitespace_index = ParserHelper.collect_backwards_while_spaces(
            detabified_original_line_prefix, -1
        )
        _, art_start_whitespace_index = ParserHelper.collect_backwards_while_spaces(
            adjusted_removed_text, -1
        )
        detabified_original_line_prefix_without_trailing_spaces = (
            detabified_original_line_prefix[:dolp_start_whitespace_index]
        )
        adjusted_removed_text_without_trailing_spaces = adjusted_removed_text[
            :art_start_whitespace_index
        ]
        # if detabified_original_line_prefix_without_trailing_spaces == adjusted_removed_text_without_trailing_spaces:
        #     return adjusted_removed_text_without_trailing_spaces, None
        assert (
            detabified_original_line_prefix_without_trailing_spaces
            == adjusted_removed_text_without_trailing_spaces
        )
        return adjusted_removed_text_without_trailing_spaces

    @staticmethod
    def __adjust_1_with_tab_loop(
        original_line_prefix: str,
        detabified_original_line_prefix: str,
        adjusted_removed_text: str,
    ) -> Optional[str]:
        original_line_prefix_without_trailing_tab = original_line_prefix[:-1]
        prefix_detabified_original_line_prefix = detabified_original_line_prefix[
            : len(original_line_prefix_without_trailing_tab)
        ]
        if (
            original_line_prefix_without_trailing_tab
            == prefix_detabified_original_line_prefix
        ):
            return BlockQuoteNonFencedHelper.__adjust_1_with_tab_part_1(
                detabified_original_line_prefix, adjusted_removed_text
            )
        return None

    @staticmethod
    def __adjust_1_with_tab(
        original_line: str, adjusted_removed_text: str
    ) -> Tuple[str, Optional[str]]:
        tabbed_removed_text = None

        if "\t" in original_line:
            POGGER.debug("original_line>>:$:<", original_line)
            detabified_original_line = TabHelper.detabify_string(original_line)
            POGGER.debug("detabified_original_line>>:$:<", detabified_original_line)
            assert detabified_original_line.startswith(
                adjusted_removed_text
            ), "Detabbified line must start with the text that was removed."
            original_line_index = 1
            while original_line_index < len(original_line):
                original_line_prefix = original_line[:original_line_index]
                POGGER.debug("original_line_prefix>>:$:<", original_line_prefix)
                detabified_original_line_prefix = TabHelper.detabify_string(
                    original_line_prefix
                )
                POGGER.debug(
                    "detabified_original_line_prefix>>:$:<",
                    detabified_original_line_prefix,
                )
                if (
                    len(detabified_original_line_prefix) > len(adjusted_removed_text)
                    and original_line_prefix[-1] == "\t"
                    and adjusted_removed_text[-1] == " "
                ):
                    loop_value = BlockQuoteNonFencedHelper.__adjust_1_with_tab_loop(
                        original_line_prefix,
                        detabified_original_line_prefix,
                        adjusted_removed_text,
                    )
                    if loop_value is not None:
                        return loop_value, None
                if detabified_original_line_prefix == adjusted_removed_text:
                    break
                original_line_index += 1
            POGGER.debug(
                "original_line_prefix>>:$:< == detabified_original_line_prefix>>:$:<",
                original_line_prefix,
                detabified_original_line_prefix,
            )
            if (
                adjusted_removed_text != original_line_prefix
                and detabified_original_line_prefix == adjusted_removed_text
            ):
                tabbed_removed_text = original_line_prefix
        return adjusted_removed_text, tabbed_removed_text

    @staticmethod
    def __block_quote_start_adjust(
        parser_state: ParserState,
        original_start_index: int,
        container_level_tokens: List[MarkdownToken],
    ) -> int:
        POGGER.debug("container_level_tokens--$", container_level_tokens)
        if container_level_tokens:
            start_index = len(parser_state.token_stack) - 1
            while start_index and not parser_state.token_stack[start_index].is_list:
                start_index -= 1
            if start_index:
                list_token = cast(ListStackToken, parser_state.token_stack[start_index])
                POGGER.debug("token_stack[start_index]--$", list_token)
                POGGER.debug(
                    "list_token.last_new_list_token--$",
                    list_token.last_new_list_token,
                )
                if (
                    list_token.last_new_list_token
                    and list_token.last_new_list_token.line_number
                    == container_level_tokens[0].line_number
                ):
                    POGGER.debug("BOOM")
                    indent_delta = (
                        list_token.last_new_list_token.indent_level
                        - list_token.last_new_list_token.column_number
                    ) - 1
                    if list_token.is_ordered_list:
                        indent_delta -= len(
                            list_token.last_new_list_token.list_start_content
                        )
                    POGGER.debug("indent_delta=:$:", indent_delta)
                    original_start_index -= indent_delta
        return original_start_index

    @staticmethod
    def __do_block_quote_leading_spaces_adjustments_adjust_bleading_part_1(
        parser_state: ParserState,
        stack_index: int,
        block_quote_token: BlockQuoteMarkdownToken,
    ) -> None:
        previous_stack_token = parser_state.token_stack[stack_index - 1]
        if previous_stack_token.is_block_quote:
            previous_markdown_token = cast(
                BlockQuoteMarkdownToken, previous_stack_token.matching_markdown_token
            )
            assert previous_markdown_token is not None
            if (
                previous_markdown_token.line_number == block_quote_token.line_number
                and previous_markdown_token.bleading_spaces == ""
            ):
                block_quote_token.weird_kludge_three = True
            if block_quote_token.leading_text_index == 1:
                assert previous_markdown_token.bleading_spaces is not None
                split_bleading_spaces = previous_markdown_token.bleading_spaces.split(
                    "\n"
                )
                block_quote_token.weird_kludge_four = (
                    previous_markdown_token.line_number,
                    previous_markdown_token.column_number,
                    previous_markdown_token.leading_text_index - 1,
                    split_bleading_spaces[
                        previous_markdown_token.leading_text_index - 1
                    ],
                )

    @staticmethod
    def __do_block_quote_leading_spaces_adjustments_adjust_bleading_part_2(
        parser_state: ParserState,
        position_marker: PositionMarker,
        stack_index: int,
        found_bq_stack_token: BlockQuoteStackToken,
    ) -> None:
        assert parser_state.token_stack[stack_index] == found_bq_stack_token
        found_list_stack_index = 0
        for search_index in range(stack_index, 0, -1):
            if (
                parser_state.token_stack[search_index].is_list
                and not found_list_stack_index
            ):
                found_list_stack_index = search_index
        if found_list_stack_index:
            list_token = cast(
                ListStartMarkdownToken,
                parser_state.token_stack[
                    found_list_stack_index
                ].matching_markdown_token,
            )
            if position_marker.line_number != list_token.line_number:
                POGGER.debug(
                    "__do_block_quote_leading_spaces_adjustments_adjust_bleading>>list_token>>$",
                    list_token,
                )
                list_token.add_leading_spaces("")
                POGGER.debug(
                    "__do_block_quote_leading_spaces_adjustments_adjust_bleading>>list_token>>$",
                    list_token,
                )

    # pylint: disable=too-many-arguments
    @staticmethod
    def __do_block_quote_leading_spaces_adjustments_adjust_bleading(
        parser_state: ParserState,
        found_bq_stack_token: BlockQuoteStackToken,
        tabbed_removed_text: Optional[str],
        adjusted_removed_text: str,
        special_case: bool,
        is_not_blank_line: bool,
        stack_index: int,
        position_marker: PositionMarker,
        text_removed_by_container: str,
    ) -> None:
        assert (
            found_bq_stack_token.matching_markdown_token is not None
        ), "Block quote stack tokens always have matching markdown tokens."
        block_quote_token = cast(
            BlockQuoteMarkdownToken, found_bq_stack_token.matching_markdown_token
        )
        POGGER.debug("dbqlsa>>last_block_token>>$", block_quote_token)
        POGGER.debug(
            "dbqlsa>>leading_text_index>>$", block_quote_token.leading_text_index
        )
        POGGER.debug("dbqlsa>>bq>>$", block_quote_token)
        POGGER.debug("dbqlsa>>tabbed_removed_text>>$", tabbed_removed_text)

        adjusted_removed_text = BlockQuoteNonFencedHelper.__check_for_kludge(
            parser_state,
            block_quote_token,
            stack_index,
            text_removed_by_container,
            adjusted_removed_text,
        )

        if not block_quote_token.weird_kludge_five:
            BlockQuoteNonFencedHelper.__do_block_quote_leading_spaces_adjustments_adjust_bleading_kludge(
                parser_state, block_quote_token, stack_index, adjusted_removed_text
            )
        POGGER.debug(
            "__do_block_quote_leading_spaces_adjustments_adjust_bleading>>block_token>>$",
            block_quote_token,
        )
        block_quote_token.add_bleading_spaces(
            adjusted_removed_text,
            special_case,
            tabbed_removed_text,
        )
        POGGER.debug(
            "__do_block_quote_leading_spaces_adjustments_adjust_bleading>>block_token>>$",
            block_quote_token,
        )

        # This checks to see if, when the first line of a block quote is encountered, if the
        # inner block quote exists and its bleading spaces are blank. If so, then the current
        # block quote was arrived at through group of block quotes together, not on building
        # on the inner block quote. If that was the case, the inner block quote would have at
        # least the bleading space from processing that block quote.  This does not mean anything
        # to the parser, but when reconstructing the Markdown, this is an important distinction.
        block_quote_token.leading_text_index += 1
        if stack_index > 1:
            BlockQuoteNonFencedHelper.__do_block_quote_leading_spaces_adjustments_adjust_bleading_part_1(
                parser_state, stack_index, block_quote_token
            )

        POGGER.debug("dbqlsa>>last_block_token>>$", block_quote_token)
        POGGER.debug(
            "dbqlsa>>leading_text_index>>$", block_quote_token.leading_text_index
        )
        if not is_not_blank_line:
            BlockQuoteNonFencedHelper.__do_block_quote_leading_spaces_adjustments_adjust_bleading_part_2(
                parser_state, position_marker, stack_index, found_bq_stack_token
            )

        POGGER.debug("__hbqs>>bq>>$", block_quote_token)

    # pylint: enable=too-many-arguments

    @staticmethod
    def __do_block_quote_leading_spaces_adjustments_adjust_bleading_kludge(
        parser_state: ParserState,
        block_quote_token: BlockQuoteMarkdownToken,
        stack_index: int,
        adjusted_removed_text: str,
    ) -> None:

        assert block_quote_token.leading_text_index == 0
        search_index = stack_index - 1
        while (
            search_index > 0
            and not parser_state.token_stack[search_index].is_block_quote
        ):
            search_index -= 1
        if search_index:  # and search_index + 1 == stack_index:
            search_token = parser_state.token_stack[
                search_index
            ].matching_markdown_token
            assert search_token is not None
            found_token = parser_state.token_stack[search_index].matching_markdown_token
            assert found_token is not None
            bq_token = cast(BlockQuoteMarkdownToken, found_token)
            assert bq_token.bleading_spaces is not None
            split_spaces = bq_token.bleading_spaces.split("\n")
            if len(split_spaces) > 1:
                last_split_space = split_spaces[-1]
                if (
                    adjusted_removed_text != last_split_space
                    and adjusted_removed_text.startswith(last_split_space)
                ):
                    block_quote_token.weird_kludge_six = True

    @staticmethod
    def __check_for_kludge(
        parser_state: ParserState,
        block_quote_token: BlockQuoteMarkdownToken,
        stack_index: int,
        text_removed_by_container: str,
        adjusted_removed_text: str,
    ) -> str:
        if not block_quote_token.bleading_spaces and stack_index > 3:
            continue_with_adjustment = (
                parser_state.token_stack[stack_index - 1].is_list
                and parser_state.token_stack[stack_index - 2].is_list
                and parser_state.token_stack[stack_index - 3].is_block_quote
            )
            if continue_with_adjustment:
                lists_new_list_token = cast(
                    ListStackToken, parser_state.token_stack[stack_index - 1]
                ).last_new_list_token
                if lists_new_list_token is not None:
                    continue_with_adjustment = (
                        block_quote_token.line_number
                        != lists_new_list_token.line_number
                    )
                else:
                    bq_inner_token = parser_state.token_stack[
                        stack_index - 3
                    ].matching_markdown_token
                    assert bq_inner_token is not None
                    continue_with_adjustment = (
                        block_quote_token.line_number != bq_inner_token.line_number
                    )
            if continue_with_adjustment:
                assert parser_state.original_line_to_parse is not None
                adjusted_removed_text = parser_state.original_line_to_parse[
                    : len(text_removed_by_container)
                ]
        return adjusted_removed_text


# pylint: enable=too-few-public-methods



"""
Module to hold the request information to pass on to the handle_* functions.
"""

from dataclasses import dataclass, field
from typing import List, Optional

from pymarkdown.container_blocks.parse_block_pass_properties import (
    ParseBlockPassProperties,
)
from pymarkdown.tokens.markdown_token import MarkdownToken
from pymarkdown.tokens.paragraph_markdown_token import ParagraphMarkdownToken


# pylint: disable=too-many-instance-attributes
@dataclass(frozen=True)
class InlineRequest:
    """
    Class to hold the request information to pass on to the handle_* functions.
    """

    source_text: str
    next_index: int
    inline_blocks: List[MarkdownToken] = field(default_factory=list)
    remaining_line: Optional[str] = None
    tabified_remaining_line: Optional[str] = None
    current_string: Optional[str] = None
    current_string_unresolved: Optional[str] = None
    line_number: Optional[int] = None
    column_number: Optional[int] = None
    para_owner: Optional[ParagraphMarkdownToken] = None
    tabified_text: Optional[str] = None
    parse_properties: Optional[ParseBlockPassProperties] = None
    last_container_token: Optional[MarkdownToken] = None
    whitespace_to_recombine: Optional[str] = None
    is_in_table: bool = False
    para_space: Optional[str] = None


# pylint: enable=too-many-instance-attributes



"""
Module to provide for a context with which tokens can be modified.
"""

from abc import abstractproperty


class PluginModifyContext:
    """
    Class to provide for a context with which tokens can be modified.
    """

    # pylint: disable=deprecated-decorator
    @abstractproperty
    def is_during_line_pass(self) -> bool:
        """
        Report on whether fix mode is currently going through a line pass.
        """

    # pylint: enable=deprecated-decorator

    # pylint: disable=deprecated-decorator
    @abstractproperty
    def in_fix_mode(self) -> bool:
        """
        Report on whether the application is in fix mode.x
        """

    # pylint: enable=deprecated-decorator



"""
Module to provide for an encapsulation of the paragraph element.
"""

from typing import Optional, Union, cast

from typing_extensions import override

from pymarkdown.general.parser_helper import ParserHelper
from pymarkdown.general.parser_logger import ParserLogger
from pymarkdown.general.position_marker import PositionMarker
from pymarkdown.tokens.leaf_markdown_token import LeafMarkdownToken
from pymarkdown.tokens.markdown_token import EndMarkdownToken, MarkdownToken
from pymarkdown.transform_gfm.transform_state import TransformState
from pymarkdown.transform_markdown.markdown_transform_context import (
    MarkdownTransformContext,
    RegisterHtmlTransformHandlersProtocol,
    RegisterMarkdownTransformHandlersProtocol,
)


class ParagraphMarkdownToken(LeafMarkdownToken):
    """
    Class to provide for an encapsulation of the paragraph element.
    """

    def __init__(
        self, extracted_whitespace: str, position_marker: PositionMarker
    ) -> None:
        self.__extracted_whitespace: str = extracted_whitespace
        self.__final_whitespace, self.rehydrate_index = (
            "",
            0,
        )
        LeafMarkdownToken.__init__(
            self,
            MarkdownToken._token_paragraph,
            "",
            position_marker=position_marker,
            requires_end_token=True,
        )
        self.__compose_extra_data_field()

    # pylint: disable=protected-access
    @staticmethod
    def get_markdown_token_type() -> str:
        """
        Get the type of markdown token for rehydration purposes.
        """
        return MarkdownToken._token_paragraph

    # pylint: enable=protected-access

    @property
    def extracted_whitespace(self) -> str:
        """
        Returns any whitespace that was extracted before the processing of this element occurred.
        """
        return self.__extracted_whitespace

    @property
    def final_whitespace(self) -> str:
        """
        Returns any final whitespace at the end of the paragraph that was removed.
        """
        return self.__final_whitespace

    def __compose_extra_data_field(self) -> None:
        """
        Compose the object's self.extra_data field from the local object's variables.
        """

        self._set_extra_data(
            f"{self.__extracted_whitespace}{MarkdownToken.extra_data_separator}{self.__final_whitespace}"
            if self.final_whitespace
            else self.__extracted_whitespace
        )

    def add_whitespace(self, whitespace_to_add: str) -> None:
        """
        Add extra whitespace to the end of the current paragraph.  Should only be
        used when combining text blocks in a paragraph.
        """

        self.__extracted_whitespace = (
            f"{self.__extracted_whitespace}{whitespace_to_add}"
        )
        self.__compose_extra_data_field()

    def set_final_whitespace(self, whitespace_to_set: str) -> None:
        """
        Set the final whitespace for the paragraph. That is any whitespace at the very
        end of the paragraph, removed to prevent hard lines at the end.
        """

        self.__final_whitespace = whitespace_to_set
        self.__compose_extra_data_field()

    def register_for_markdown_transform(
        self, registration_function: RegisterMarkdownTransformHandlersProtocol
    ) -> None:
        """
        Register any rehydration handlers for leaf markdown tokens.
        """
        registration_function(
            ParagraphMarkdownToken,
            ParagraphMarkdownToken.__rehydrate_paragraph,
            ParagraphMarkdownToken.__rehydrate_paragraph_end,
        )

    @staticmethod
    def __rehydrate_paragraph(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the paragraph block from the token.
        """
        _ = previous_token

        context.block_stack.append(current_token)

        current_paragraph_token = cast(ParagraphMarkdownToken, current_token)
        current_paragraph_token.rehydrate_index = 0
        extracted_whitespace = current_paragraph_token.extracted_whitespace
        if ParserHelper.newline_character in extracted_whitespace:
            line_end_index = extracted_whitespace.index(ParserHelper.newline_character)
            extracted_whitespace = extracted_whitespace[:line_end_index]
        return f"{ParserLogger.start_range_sequence}{ParserHelper.resolve_all_from_text(extracted_whitespace)}"

    @staticmethod
    def __rehydrate_paragraph_end(
        context: MarkdownTransformContext,
        current_token: MarkdownToken,
        previous_token: Optional[MarkdownToken],
        next_token: Optional[MarkdownToken],
    ) -> str:
        """
        Rehydrate the end of the paragraph block from the token.
        """
        _ = (previous_token, next_token)

        top_stack_token = cast(ParagraphMarkdownToken, context.block_stack[-1])
        del context.block_stack[-1]

        current_end_token = cast(EndMarkdownToken, current_token)
        current_start_token = cast(
            ParagraphMarkdownToken, current_end_token.start_markdown_token
        )

        rehydrate_index, expected_rehydrate_index = (
            current_start_token.rehydrate_index,
            ParserHelper.count_newlines_in_text(
                current_start_token.extracted_whitespace
            ),
        )
        assert (
            rehydrate_index == expected_rehydrate_index
        ), "Rehydrate index must match up at end of paragraph."
        return f"{top_stack_token.final_whitespace}{ParserLogger.end_range_sequence}{ParserHelper.newline_character}"

    @staticmethod
    def register_for_html_transform(
        register_handlers: RegisterHtmlTransformHandlersProtocol,
    ) -> None:
        """
        Register any functions required to generate HTML from the tokens.
        """
        register_handlers(
            ParagraphMarkdownToken,
            ParagraphMarkdownToken.__handle_start_paragraph_token,
            ParagraphMarkdownToken.__handle_end_paragraph_token,
        )

    @staticmethod
    def __handle_start_paragraph_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token
        token_parts = [output_html]
        if output_html and output_html[-1] != ParserHelper.newline_character:
            token_parts.append(ParserHelper.newline_character)
        if transform_state.is_in_loose_list:
            token_parts.append("<p>")
        return "".join(token_parts)

    @staticmethod
    def __handle_end_paragraph_token(
        output_html: str,
        next_token: MarkdownToken,
        transform_state: TransformState,
    ) -> str:
        _ = next_token

        return (
            f"{output_html}</p>{ParserHelper.newline_character}"
            if transform_state.is_in_loose_list
            else output_html
        )

    @override
    def _modify_token(self, field_name: str, field_value: Union[str, int]) -> bool:
        if field_name == "extracted_whitespace" and isinstance(field_value, str):
            self.__extracted_whitespace = field_value
            self.__compose_extra_data_field()
            return True
        return False



"""
Module to provide some extra support to the ApplicationPropertiesFacade class.
"""

from typing import Any, Callable, Optional

from application_properties import ApplicationPropertiesFacade


class MyApplicationPropertiesFacade:
    """
    Class to provide extra functionality to the ApplicationPropertiesFacade class.
    If proven, the ultimate goal is to include this support in `application_properties`.
    """

    def __init__(self, properties_facade: ApplicationPropertiesFacade):
        self.__properties_facade = properties_facade

    def get_boolean_property_with_default(
        self,
        property_name: str,
        default_value: bool,
        is_required: bool = False,
    ) -> bool:
        """
        Get a boolean property from the configuration.
        """
        intermediate_result = self.__properties_facade.get_boolean_property(
            property_name=property_name,
            default_value=default_value,
            is_required=is_required,
        )
        assert intermediate_result is not None
        return intermediate_result

    # pylint: disable=too-many-arguments
    def get_integer_property_with_default(
        self,
        property_name: str,
        default_value: int,
        valid_value_fn: Optional[Callable[[int], Any]] = None,
        is_required: bool = False,
        strict_mode: Optional[bool] = None,
    ) -> int:
        """
        Get an integer property from the configuration.

        TODO update docstrings for App_properties to be reflective of what they are.
        """
        intermediate_result = self.__properties_facade.get_integer_property(
            property_name=property_name,
            default_value=default_value,
            valid_value_fn=valid_value_fn,
            is_required=is_required,
            strict_mode=strict_mode,
        )
        assert intermediate_result is not None
        return intermediate_result

    # pylint: enable=too-many-arguments
    # pylint: disable=too-many-arguments
    def get_string_property_with_default(
        self,
        property_name: str,
        default_value: str,
        valid_value_fn: Optional[Callable[[str], Any]] = None,
        is_required: bool = False,
        strict_mode: Optional[bool] = None,
    ) -> str:
        """
        Get a string property from the configuration.
        """
        intermediate_result = self.__properties_facade.get_string_property(
            property_name=property_name,
            default_value=default_value,
            valid_value_fn=valid_value_fn,
            is_required=is_required,
            strict_mode=strict_mode,
        )
        assert intermediate_result is not None
        return intermediate_result

    # pylint: enable=too-many-arguments

    def get_boolean_property(
        self,
        property_name: str,
        default_value: Optional[bool] = None,
        is_required: bool = False,
    ) -> Optional[bool]:
        """
        Get a boolean property from the configuration.
        """
        return self.__properties_facade.get_boolean_property(
            property_name=property_name,
            default_value=default_value,
            is_required=is_required,
        )

    # pylint: disable=too-many-arguments
    def get_integer_property(
        self,
        property_name: str,
        default_value: Optional[int] = None,
        valid_value_fn: Optional[Callable[[int], Any]] = None,
        is_required: bool = False,
        strict_mode: Optional[bool] = None,
    ) -> Optional[int]:
        """
        Get an integer property from the configuration.
        """
        return self.__properties_facade.get_integer_property(
            property_name=property_name,
            default_value=default_value,
            valid_value_fn=valid_value_fn,
            is_required=is_required,
            strict_mode=strict_mode,
        )

    # pylint: enable=too-many-arguments

    # pylint: disable=too-many-arguments
    def get_string_property(
        self,
        property_name: str,
        default_value: Optional[str] = None,
        valid_value_fn: Optional[Callable[[str], Any]] = None,
        is_required: bool = False,
        strict_mode: Optional[bool] = None,
    ) -> Optional[str]:
        """
        Get a string property from the configuration.
        """
        return self.__properties_facade.get_string_property(
            property_name=property_name,
            default_value=default_value,
            valid_value_fn=valid_value_fn,
            is_required=is_required,
            strict_mode=strict_mode,
        )

    # pylint: enable=too-many-arguments



"""
Module to hold the response from the inline handle_* functions.
"""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Optional


# pylint: disable=too-many-instance-attributes
@dataclass
class InlineResponse:
    """
    Class to hold the response from the inline handle_* functions.
    """

    new_string: Optional[str] = None
    new_string_unresolved: Optional[str] = None
    new_index: Optional[int] = None
    new_tokens: list = field(default_factory=list)  # type: ignore
    consume_rest_of_line: bool = False
    original_string: Optional[str] = None
    delta_line_number: int = 0
    delta_column_number: int = 0
    reduce_remaining_line_by: int = 0
    adj_newlines: int = 0


# pylint: enable=too-many-instance-attributes



"""
Module to implement a plugin that looks to see if the first heading in a file is
a top level heading.
"""

from typing import List, cast

from pymarkdown.plugin_manager.plugin_details import (
    PluginDetails,
    PluginDetailsV3,
    QueryConfigItem,
)
from pymarkdown.plugin_manager.plugin_scan_context import PluginScanContext
from pymarkdown.plugin_manager.rule_plugin import RulePlugin
from pymarkdown.tokens.atx_heading_markdown_token import AtxHeadingMarkdownToken
from pymarkdown.tokens.markdown_token import MarkdownToken


class RuleMd002(RulePlugin):
    """
    Class to implement a plugin that looks to see if the first heading in a file is
    a top level heading.
    """

    def __init__(self) -> None:
        super().__init__()
        self.__start_level = 0
        self.__have_seen_first_heading = False

    def get_details(self) -> PluginDetails:
        """
        Get the details for the plugin.
        """
        return PluginDetailsV3(
            plugin_name="first-heading-h1,first-header-h1",
            plugin_id="MD002",
            plugin_enabled_by_default=False,
            plugin_description="First heading of the document should be a top level heading.",
            plugin_version="0.6.1",
            plugin_url="https://pymarkdown.readthedocs.io/en/latest/plugins/rule_md002.md",
            plugin_configuration="level",
        )

    @classmethod
    def __validate_configuration_other_test_value(cls, found_value: int) -> None:
        if not 1 <= found_value <= 6:
            raise ValueError("Allowable values are between 1 and 6 (inclusive).")

    def initialize_from_config(self) -> None:
        """
        Event to allow the plugin to load configuration information.
        """
        self.__start_level = (
            self.plugin_configuration.get_integer_property_with_default(
                "level",
                1,
                valid_value_fn=self.__validate_configuration_other_test_value,
            )
        )

    def query_config(self) -> List[QueryConfigItem]:
        """
        Query to find out the configuration that the rule is using.
        """
        return [QueryConfigItem("level", self.__start_level)]

    def starting_new_file(self) -> None:
        """
        Event that the a new file to be scanned is starting.
        """
        self.__have_seen_first_heading = False

    def next_token(self, context: PluginScanContext, token: MarkdownToken) -> None:
        """
        Event that a new token is being processed.
        """
        if token.is_atx_heading or token.is_setext_heading:
            heading_token = cast(AtxHeadingMarkdownToken, token)
            hash_count = heading_token.hash_count
        else:
            hash_count = None

        if not self.__have_seen_first_heading and hash_count:
            self.__have_seen_first_heading = True
            if hash_count != self.__start_level:
                extra_data = f"Expected: h{self.__start_level}; Actual: h{hash_count}"
                self.report_next_token_error(
                    context,
                    token,
                    extra_error_information=extra_data,
                    use_original_position=token.is_setext_heading,
                )
